# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Users\Reinh\OneDrive\Dokumente\python\MzChess\propagatePawn.ui'
#
# Created by: PyQt5 UI code generator 5.15.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(220, 70)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Form.sizePolicy().hasHeightForWidth())
        Form.setSizePolicy(sizePolicy)
        Form.setMinimumSize(QtCore.QSize(220, 70))
        Form.setMaximumSize(QtCore.QSize(220, 70))
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.queenButton = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.queenButton.sizePolicy().hasHeightForWidth())
        self.queenButton.setSizePolicy(sizePolicy)
        self.queenButton.setMinimumSize(QtCore.QSize(50, 50))
        self.queenButton.setMaximumSize(QtCore.QSize(50, 50))
        font = QtGui.QFont()
        font.setPointSize(30)
        self.queenButton.setFont(font)
        self.queenButton.setObjectName("queenButton")
        self.horizontalLayout.addWidget(self.queenButton)
        self.rookButton = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.rookButton.sizePolicy().hasHeightForWidth())
        self.rookButton.setSizePolicy(sizePolicy)
        self.rookButton.setMinimumSize(QtCore.QSize(50, 50))
        self.rookButton.setMaximumSize(QtCore.QSize(50, 50))
        font = QtGui.QFont()
        font.setPointSize(30)
        self.rookButton.setFont(font)
        self.rookButton.setObjectName("rookButton")
        self.horizontalLayout.addWidget(self.rookButton)
        self.bishopButton = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bishopButton.sizePolicy().hasHeightForWidth())
        self.bishopButton.setSizePolicy(sizePolicy)
        self.bishopButton.setMinimumSize(QtCore.QSize(50, 50))
        self.bishopButton.setMaximumSize(QtCore.QSize(50, 50))
        font = QtGui.QFont()
        font.setPointSize(30)
        self.bishopButton.setFont(font)
        self.bishopButton.setObjectName("bishopButton")
        self.horizontalLayout.addWidget(self.bishopButton)
        self.knightButton = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.knightButton.sizePolicy().hasHeightForWidth())
        self.knightButton.setSizePolicy(sizePolicy)
        self.knightButton.setMinimumSize(QtCore.QSize(50, 50))
        self.knightButton.setMaximumSize(QtCore.QSize(50, 50))
        font = QtGui.QFont()
        font.setPointSize(30)
        self.knightButton.setFont(font)
        self.knightButton.setObjectName("knightButton")
        self.horizontalLayout.addWidget(self.knightButton)
        self.verticalLayout.addLayout(self.horizontalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Propagate ..."))
        self.queenButton.setText(_translate("Form", "♕"))
        self.rookButton.setText(_translate("Form", "♖"))
        self.bishopButton.setText(_translate("Form", "♗"))
        self.knightButton.setText(_translate("Form", "♘"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())



from typing import Union, List, Dict
from PyQt5 import QtCore, QtGui, QtWidgets

def treeWidgetItemPos(item : QtWidgets.QTreeWidgetItem) -> QtCore.QPoint:
  pos = item.treeWidget().visualItemRect(item).bottomLeft()
  return item.treeWidget().mapToParent(pos)

class ButtonLine(QtWidgets.QDialog):
 def __init__(self, buttonTexts : Union[List[str], Dict[str, int]],
                          hintDict : Dict[str, str] = dict(), 
                          title : str = ' ', 
                          pointSize : int = 12, 
                          parent = None) -> None:
  super(ButtonLine, self).__init__(parent)
  self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
  self.hBox = QtWidgets.QHBoxLayout()
  self.hBox.setContentsMargins(pointSize // 2, pointSize // 2, pointSize // 2,  pointSize // 2)
  self.hBox.setSpacing(0)
  font = QtGui.QFont()
  font.setPointSize(pointSize)
  self.setFont(font)
  fontMetrics = self.fontMetrics()
  self.buttonWidth = 0
  for txt in buttonTexts:
   self.buttonWidth = max(self.buttonWidth, fontMetrics.size(QtCore.Qt.TextSingleLine, txt).width())
  self.buttonWidth = self.buttonWidth + pointSize
  self.buttonHeight = 2 * pointSize
  self.buttonTextDict = dict()
  for ID, txt in enumerate(buttonTexts):
   if isinstance(buttonTexts, list):
    self.buttonTextDict[txt] = ID
   if txt in hintDict:
    hint = hintDict[txt]
   else:
    hint = None
   self._addButton(txt, hint)
  if isinstance(buttonTexts, dict):
   self.buttonTextDict = buttonTexts
  self._setSizePolicy(self)
  self.resize(len(buttonTexts) * self.buttonWidth, self.buttonHeight)
  self.setLayout(self.hBox)
  self.setWindowTitle(title)
  
 def setFocus(self, buttonText : str) -> None:
  buttonID = self.buttonTextDict[buttonText]
  self.children()[buttonID].setFocus()
 
 def _setSizePolicy(self, widget : QtWidgets.QWidget) -> None:
  sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
  sizePolicy.setHorizontalStretch(0)
  sizePolicy.setVerticalStretch(0)
  sizePolicy.setHeightForWidth(widget.sizePolicy().hasHeightForWidth())
  widget.setSizePolicy(sizePolicy)
 
 def _addButton(self, txt : str,  hint : str) -> None:
  pushButton = QtWidgets.QPushButton(self)
  pushButton.setMinimumSize(QtCore.QSize(self.buttonWidth, self.buttonHeight))
  pushButton.setMaximumSize(QtCore.QSize(self.buttonWidth, self.buttonHeight))
  pushButton.setAutoDefault(True)
  pushButton.setText(txt)
  if hint is not None:
   pushButton.setToolTip(hint)
  pushButton.clicked.connect(self.on_clicked)
  self.hBox.addWidget(pushButton)
  
 @QtCore.pyqtSlot()
 def on_clicked(self):
  sendingItem = self.sender()
  self.done(self.buttonTextDict[sendingItem.text()])
  
class TextEdit(QtWidgets.QDialog):
 def __init__(self, title : str, pointSize : int = 10, parent = None) -> None:
  super(TextEdit, self).__init__(parent)
  self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
  font = QtGui.QFont()
  font.setPointSize(pointSize)
  self.setFont(font)

  self.verticalLayout = QtWidgets.QVBoxLayout(self)
  self.edit = QtWidgets.QPlainTextEdit(self)
  self.buttonBox = QtWidgets.QDialogButtonBox(self)
  self.buttonBox.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
  self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
  self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)

  self.verticalLayout.addWidget(self.edit)
  self.verticalLayout.addWidget(self.buttonBox)

  self.buttonBox.accepted.connect(self.accept)
  self.buttonBox.rejected.connect(self.reject)
  self.setWindowTitle(title)
  
 def setPos(self, masterPos : QtCore.QPoint) -> None:
  self.move(masterPos)
  
 def setText(self, txt : str) -> None:
  self.edit.setPlainText(txt)

 def text(self) -> str:
  return self.edit.toPlainText()

 def accept(self) -> str:
  return self.done(True)

 def reject(self) -> str:
  return self.done(False)

class ItemSelector(QtWidgets.QDialog):
 def __init__(self, title : str, pointSize : int = 10, parent = None) -> None:
  super(ItemSelector, self).__init__(parent)
  self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
  font = QtGui.QFont()
  font.setPointSize(pointSize)
  self.setFont(font)

  self.verticalLayout = QtWidgets.QVBoxLayout(self)
  self.listWidget = QtWidgets.QListWidget(self)
  self.buttonBox = QtWidgets.QDialogButtonBox(self)
  self.buttonBox.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
  self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
  self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)

  self.verticalLayout.addWidget(self.listWidget)
  self.verticalLayout.addWidget(self.buttonBox)

  self.buttonBox.accepted.connect(self.accept)
  self.buttonBox.rejected.connect(self.reject)
  self.setWindowTitle(title)
  
 def setContent(self, allItems : List[str], selectedItems : List[str] = list()) -> None:
  self.listWidget.clear()
  for itemTxt in sorted(allItems):
   item = QtWidgets.QListWidgetItem(itemTxt, self.listWidget)
   item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
   if itemTxt in selectedItems:
    item.setCheckState(QtCore.Qt.Checked)
   else:
    item.setCheckState(QtCore.Qt.Unchecked)

 def selectedItems(self) -> List[str]:
  selItems = list()
  for item in self.listWidget.findItems('*', QtCore.Qt.MatchWildcard):
   if item.checkState() == QtCore.Qt.Checked:
    selItems.append(item.text())
  return selItems

 def accept(self) -> str:
  return self.done(True)

 def reject(self) -> str:
  return self.done(False)


if __name__ == "__main__":
 import chess
 
 app = QtWidgets.QApplication([])
 if True:
  from enum import IntEnum, unique
  @unique
  class KeyType(IntEnum):
   STR = 0
   INT = 1
   URL = 2
   LABEL = 3
   DATE = 4
   TIME = 5
   EVENT = 6
   SITE = 7
   PLAYER = 8
   TITLE = 9
  stdKey2ValueTypeDict = {
   # seven tag roster
   "Event" : ("?", KeyType.EVENT),
   "Site" : ("?", KeyType.SITE),
   "Date" : ("????.??.??", KeyType.DATE),
   "Round" : ("?", KeyType.STR),
   "White" : ("?", KeyType.PLAYER),
   "Black" : ("?", KeyType.PLAYER),
   "Result" : ("*", KeyType.STR), 
   # player related information 
   "WhiteTitle" : ("", KeyType.TITLE), 
   "BlackTitle" : ("", KeyType.TITLE), 
   "WhiteElo" : ("", KeyType.INT)}
  itemSelect = ItemSelector('Header Elements ...', pointSize = 10)
  itemSelect.setContent(stdKey2ValueTypeDict, ["WhiteElo", "Result"])
  index = itemSelect.exec()
  print('selected = {}'.format(itemSelect.selectedItems()))
 elif True:
  annotationSymbols = ['', '!', '?', '!!', '??', '!?', '?!']
  positionSymbols =  ['','=', '~', '+=', '=+', '+/-', '-/+', '+-','-+', '+--', '-++']
  symbolDescription =  { ''    : 'null annotation',
                              '!'     : 'good move', 
                              '?'     : 'poor move', 
                              '!!'   : 'brilliant move',
                              '??'   : 'blunder',
                              '!?'   : 'intesting move',
                              '?!'   : 'dubious move', 
                              '='     : 'even position', 
                              '~'     : 'unclear position', 
                              '+='   : 'slight advantage for white', 
                              '=+'   : 'slight advantage for black',
                              '+/-' : 'moderate advantage for white', 
                              '-/+' : 'moderate advantage for black', 
                              '+-'   : 'decisive advantage for white',
                              '-+'   : 'decisive advantage for white',
                              '+--' : 'white should resign',
                              '-++' : 'black should resign' }
  chessFigures = { '♕' :  chess.QUEEN, '♖' :  chess.ROOK,  '♗' :  chess.BISHOP,  '♘' :  chess.KNIGHT}

  # buttonLine = ButtonLine(annotationSymbols, hintDict = symbolDescription, pointSize = 12)
  buttonLine = ButtonLine(positionSymbols, hintDict = symbolDescription, pointSize = 12)
  # buttonLine = ButtonLine(chessFigures, pointSize = 30)
  index = buttonLine.exec()
 else:
  myText = '''[Event "Whatever"]
[Site "Wherever"]
[Date "1970.01.01"]
[Round "5"]
[White "Foo"]
[Black "Bla"]
[Result "*"]
1. d4 { [%eval +0.16] } 

1... d5 *
[Event "Whatever"]
[Site "?"]
[Date "1970.01.01"]
[Round "13"]
[White "Foo"]
[Black "Bla"]
[Result "*"]

1. d4 { [%eval +0.16] } 1... c5 { [%eval +0.52] (Benoni-Verteidigung) } *
  '''
  textEdit = TextEdit('Comment ...', pointSize = 10)
  textEdit.setText(myText)
  index = textEdit.exec()
 print(index)
 #sys.exit(app.exec_())




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Users\Reinh\OneDrive\Dokumente\python\MzChess\MzChess\AboutDialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.resize(268, 204)
        AboutDialog.setSizeGripEnabled(False)
        self.verticalLayout = QtWidgets.QVBoxLayout(AboutDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.pgmLabel = QtWidgets.QLabel(AboutDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pgmLabel.sizePolicy().hasHeightForWidth())
        self.pgmLabel.setSizePolicy(sizePolicy)
        self.pgmLabel.setMinimumSize(QtCore.QSize(250, 0))
        font = QtGui.QFont()
        font.setPointSize(16)
        self.pgmLabel.setFont(font)
        self.pgmLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.pgmLabel.setObjectName("pgmLabel")
        self.verticalLayout.addWidget(self.pgmLabel)
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setObjectName("formLayout")
        self.label_2 = QtWidgets.QLabel(AboutDialog)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_2)
        self.label_3 = QtWidgets.QLabel(AboutDialog)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.label_3)
        self.label_4 = QtWidgets.QLabel(AboutDialog)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_4)
        self.shoeboxVersionLabel = QtWidgets.QLabel(AboutDialog)
        self.shoeboxVersionLabel.setObjectName("shoeboxVersionLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.shoeboxVersionLabel)
        self.label_6 = QtWidgets.QLabel(AboutDialog)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.shoeboxDateLabel = QtWidgets.QLabel(AboutDialog)
        self.shoeboxDateLabel.setObjectName("shoeboxDateLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.shoeboxDateLabel)
        self.label_7 = QtWidgets.QLabel(AboutDialog)
        self.label_7.setObjectName("label_7")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.label_7)
        self.pythonVersionLabel = QtWidgets.QLabel(AboutDialog)
        self.pythonVersionLabel.setObjectName("pythonVersionLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.pythonVersionLabel)
        self.label_11 = QtWidgets.QLabel(AboutDialog)
        self.label_11.setObjectName("label_11")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.label_11)
        self.qtVersionLabel = QtWidgets.QLabel(AboutDialog)
        self.qtVersionLabel.setObjectName("qtVersionLabel")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.qtVersionLabel)
        self.label_5 = QtWidgets.QLabel(AboutDialog)
        self.label_5.setObjectName("label_5")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.label_5)
        self.chessVersionLabel = QtWidgets.QLabel(AboutDialog)
        self.chessVersionLabel.setObjectName("chessVersionLabel")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.chessVersionLabel)
        self.label = QtWidgets.QLabel(AboutDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.label)
        self.pyqt5VersionLabel = QtWidgets.QLabel(AboutDialog)
        self.pyqt5VersionLabel.setObjectName("pyqt5VersionLabel")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.pyqt5VersionLabel)
        self.label_8 = QtWidgets.QLabel(AboutDialog)
        self.label_8.setObjectName("label_8")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.label_8)
        self.plyVersionLabel = QtWidgets.QLabel(AboutDialog)
        self.plyVersionLabel.setObjectName("plyVersionLabel")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.plyVersionLabel)
        self.verticalLayout.addLayout(self.formLayout)

        self.retranslateUi(AboutDialog)
        QtCore.QMetaObject.connectSlotsByName(AboutDialog)

    def retranslateUi(self, AboutDialog):
        _translate = QtCore.QCoreApplication.translate
        AboutDialog.setWindowTitle(_translate("AboutDialog", "About ..."))
        self.pgmLabel.setText(_translate("AboutDialog", "Mz Chess GUI"))
        self.label_2.setText(_translate("AboutDialog", "Author:"))
        self.label_3.setText(_translate("AboutDialog", "Reinhard Maerz"))
        self.label_4.setText(_translate("AboutDialog", "Version:"))
        self.shoeboxVersionLabel.setText(_translate("AboutDialog", "V?.?"))
        self.label_6.setText(_translate("AboutDialog", "Version-date:"))
        self.shoeboxDateLabel.setText(_translate("AboutDialog", "Apr 99, 2019"))
        self.label_7.setText(_translate("AboutDialog", "Python version:"))
        self.pythonVersionLabel.setText(_translate("AboutDialog", "V?.?.?"))
        self.label_11.setText(_translate("AboutDialog", "Qt version:"))
        self.qtVersionLabel.setText(_translate("AboutDialog", "V?.?.?"))
        self.label_5.setText(_translate("AboutDialog", "chess version"))
        self.chessVersionLabel.setText(_translate("AboutDialog", "V?.?"))
        self.label.setText(_translate("AboutDialog", "PyQT5 version"))
        self.pyqt5VersionLabel.setText(_translate("AboutDialog", "V?.?.?"))
        self.label_8.setText(_translate("AboutDialog", "ply"))
        self.plyVersionLabel.setText(_translate("AboutDialog", "V?.?"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    AboutDialog = QtWidgets.QDialog()
    ui = Ui_AboutDialog()
    ui.setupUi(AboutDialog)
    AboutDialog.show()
    sys.exit(app.exec_())



'''A new-Parser for Portable Game Notation (`PGN`_) files for `chess.pgn`_ items
 
.. _chess.pgn: https://pypi.org/project/chess
.. _PGN: https://github.com/fsmosca/PGN-Standard
'''

from typing import Optional, Type, Callable, List, TextIO
import os, os.path
import re

import chess.pgn
import ply.lex

class PGNLexer(object):
 '''A Lexer for Portable Game Notation (PGN) files (see `ply`_)
tracking line number, gameID, position (in game) 

:param bufsize: size of the token buffer (min: 4096). The buffer is refilled at bufsize/4 
:param debug: run the lexer in debug mode
:param kwargs: other keyword arguments of ply.lex.lex 
 
.. _ply: https://ply.readthedocs.io/en/latest/
 '''
 states = ( ('comment','exclusive'),  ('tag', 'exclusive') )
 tokens = ( 'TAGNAME', 'TAGVALUE', 'ENDOFGAME', 'MOVENUMBER', 'SANPLY', 
                'LPAREN', 'RPAREN', 'LBRACE', 'RBRACE', 'LBRACK', 'RBRACK', 'NAG',
                'COMMENT', 'LINECOMMENT')
 nagDict = { '?' : 2,  '??' : 4, '?!' : 6, '!' : 1,  '!!' : 3, '!?' : 5, 
                  '+=' :  14, '=+' : 15, '+/-' : 16, '+-' : 18, '+--' : 20,
                  '-/+' : 17, '-+' : 19, '--+' : 21,  '=' : 10, '~' : 13, 'N' : 146, 'D' : 220}
 
 def __init__(self, bufsize : int = 2147483647, debug : bool = False, **kwargs) -> None:
  assert bufsize >= 4096
  self.bufsize = bufsize
  self.lexer = ply.lex.lex(module = self, debug = debug, **kwargs)
  self.gameID = 0
  self.data = None
  self.f = None

 def newGame(self, f : TextIO) -> ply.lex.LexToken:
  '''Prepared the lexer to deliver tokens for a game 
  
:param f: file handle opened in text mode 
:returns: the first token of the game
  '''
  if f.tell() == 0 or f is not self.f:
   lexer.lineno = 1
   lexer.gameID = 0
   self.f = f
   self.data = self.f.read(self.bufsize).lstrip("\ufeff")
  self.gameID += 1
  self.lexer.input(self.data)
  return self.token()
  
 def endGame(self) -> None:
  'Ends running through a new game'
  self.data = self.data[self.lexer.lexpos:]
  self.lexer.input(self.data)
  self._loadBuffer()

 def _loadBuffer(self):
  if len(self.data) == self.bufsize and len(self.data) - self.lexer.lexpos < self.bufsize // 4:
   self.data += self.f.read(self.bufsize)
   self.data = self.data[self.lexer.lexpos:]
   self.lexer.input(self.data)

 def token(self) -> ply.lex.LexToken:
  self._loadBuffer()
  return self.lexer.token()
  
 def dumps(self, tok : Optional[ply.lex.LexToken] = None, notify : Optional[Callable[[str], None]] = None) -> None:
  msgList = list()
  if tok is None:
   tok = self.token()
  while tok is not None:
   if notify is not None:
    notify('({}({}): {}/{})'.format(tok.type, tok.value, tok.lineno, tok.lexpos))
   msgList.append('({}({}): {}/{})'.format(tok.type, tok.value, tok.lineno, tok.lexpos))
   tok = self.token()
  return '\n'.join(msgList)
  
 def endOfGamePattern(self) -> re.Pattern:
  '''Returns the compiled *t_ENDOFGAME* pattern
  
:returns: compiled *t_ENDOFGAME* pattern
  '''
  return re.compile(self.t_ENDOFGAME) 

 def _raise_error(self, rootCause : str,  t : ply.lex, warn = False) -> None:
  if isinstance(t.value, str):
   txt = t.value[:20]
  else:
   txt = repr(t.value)
  if warn:
   print("LEXER/{}: '{} ...' @ Line: {}, Game {}, Pos: {}".format(rootCause, txt, t.lineno, self.gameID,  t.lexpos))
  else:
   raise SyntaxError("LEXER/{}: '{} ...' @ Line: {}, Game {}, Pos: {}".format(rootCause, txt, t.lineno, self.gameID,  t.lexpos))
 
 # ---------------------------------------------------------------------------------- 
 
 # Completely ignored characters
 t_ignore = ' \t\x0c'
 t_tag_ignore = ' \t\x0c'
 t_comment_ignore = '\x0c'
 t_ENDOFGAME = r'(1\-0|0\-1|1/2\-1/2|\*)'

 # Newlines
 def t_NEWLINE(self, t):
  r'\n+'
  t.lexer.lineno += t.value.count("\n")

 def t_tag_NEWLINE(self, t):
  r'\n+'
  t.lexer.lineno += t.value.count("\n")

 t_tag_TAGNAME = r'[A-Z][A-Za-z0-9]*'
 t_LPAREN = r'\('
 t_RPAREN = r'\)'

 def t_tag_TAGVALUE(self, t): 
  r'\"(\\\"|[^"])*\"'
  t.value = t.value[1:-1]
  return t

 def t_SANPLY(self, t):
  r'(([NBKRQ]?[a-h]?[1-8]?)?[\:x]?[a-h][1-8](=[QRBN])?[+#]?|O-O(-O)?|0-0(-0)?|0{4}|@{4}|Z0|--(?![+\-]))'
  t.value = t.value.replace('x', '').replace(':', '')
  t.value = t.value.replace('+', '').replace('#', '')
  if t.value == '@@@@' or t.value == 'Z0':
   t.value = '0000'
  return t

 def t_LBRACK(self, t):
  r'\['
  t.lexer.begin('tag')
 
 def t_tag_RBRACK(self, t):
  r'\]'
  t.lexer.begin('INITIAL')

 def  t_MOVENUMBER(self, t):
  r'([1-9][0-9]*\.*|\.+)(?!-)'
  pass
  
 def t_NAG(self, t):
  r'(\$[1-9][0-9]{0,2}(?![0-9])|(--\+|\+--|-/\+|\+/-|-\+|-\+|\+-|-\+|\+=|=\+)(?![=+\-])|(\?[?!]|![?!])(?![?!])|[?!=~DN](?![?!=~DN]))'
  if t.value[0] != '$':
   if t.value not in self.nagDict:
    self._raise_error("Illegal NAG symbol", t, warn = True)
    t.value = 0
   else:
    t.value = self.nagDict[t.value]
  else:
   t.value = int(t.value[1:])
   if t.value >= 140:
    self._raise_error("Illegal NAG string", t, warn = True)
    t.value = 0
  return t  
  
 def t_comment_COMMENT(self, t):
  r'[^{}]+'
  t.lexer.lineno += t.value.count("\n")
  return t

 def t_LBRACE(self, t):
  r'\{'
  t.lexer.begin('comment')
 
 def t_comment_RBRACE(self, t):
  r'\}'
  t.lexer.begin('INITIAL')
 
 def t_tag_LINECOMMENT(self, t):
  r';[^\n]*'
  pass

 def t_LINECOMMENT(self, t):
  r';[^\n]*'
  pass

  # -----------------------------------------------------------------------

 def  t_TAGNAME (self, t):
  r'\['
  self._raise_error("Illegal tagname", t)

 def t_TAGVALUE(self, t):
  r'\]'
  self._raise_error("Illegal tagname", t)

 def t_RBRACK(self, t):
  r'\]'
  self._raise_error("Unexpected ]")

 def t_COMMENT(self, t):
  r'\{'
  self._raise_error("Illegal comment string", t)
 
 def t_RBRACE(self, t):
  r'\}'
  self._raise_error("Unexpected }")

 def t_error(self, t):
  self._raise_error("Illegal character", t)

 def t_tag_error(self, t):
  self._raise_error("Illegal tag character", t)

 def t_comment_error(self, t):
  self._raise_error("Illegal comment character", t)

# ==================================================================

lexer : PGNLexer = PGNLexer()
'''The global instance of the lexer'''

def _checkToken(tok : ply.lex.LexToken, expectedType : str) -> None:
 global lexer
 if tok is None:
  raise SyntaxError('Parser: {} marker missing'.format(expectedType))
 elif tok.type != expectedType:
  raise SyntaxError('Parser: Unexpected token "{}" = {} ... (expected "{}") @ Line {}, Game {}, Pos {}'.format(
                            tok.type, str(tok.value)[:10], expectedType, tok.lineno, lexer.gameID, tok.lexpos))

def _skipGameBody() -> None:
 global lexer
 data = lexer.lexer.lexdata[lexer.lexer.lexpos:]
 actPos = 0

 if len(data) == 0:
  return lexer.lexer.lexpos
 
 endOfGameID = 0
 brackID = -1

 patternList = list()
 patternList.append(lexer.endOfGamePattern())
 patternList.append(re.compile(r'\{[^{}]*\}'))
 patternList.append(re.compile(r';[^\n]*\n'))

 matchDict = dict()
 for id in range(3):  
  match = patternList[id].search(data)
  if match is not None:
   span = match.span()
   # print('Setup: {}/{}: {}'.format(id, actPos + span[1], data[actPos + span[0]:actPos + span[1]]))
   matchDict[actPos + span[0]] = (id, actPos + span[1])

 newBrackPos = data[actPos:].find('[')
 if newBrackPos >= 0:
  brackPos = actPos + newBrackPos
 else:
  brackPos = 2147483647
 while len(matchDict) > 0:
  minPos = sorted(matchDict)[0]
  id, endPos = matchDict[minPos]
  if brackPos < minPos:
   id = brackID
   endPos = minPos +1
   break
  if id == endOfGameID:
   break
  del matchDict[minPos]
  idList = [id]
  keys = list(matchDict.keys())
  for pos in keys:
   if pos < endPos:
    id, _ = matchDict[pos]
    del matchDict[pos]
    idList.append(id)
  actPos = endPos
  for id in idList:
   match = patternList[id].search(data[actPos:])
   if match is not None:
    span = match.span()
    # print('{}/{}: {}'.format(id, actPos + span[1], data[actPos + span[0]:actPos + span[1]]))
    matchDict[actPos + span[0]] = (id, actPos + span[1])
  newBrackPos = data[actPos:].find('[')
  if newBrackPos >= 0:
   brackPos = actPos + newBrackPos
  else:
   brackPos = 2147483647
    
 lexer.lexer.lineno += data[:endPos].count('\n')
 lexer.lexer.lexpos += endPos - (id == brackID)
 if id == endOfGameID:
  return None
 if len(matchDict) == 0:
  raise SyntaxError("SkipBody: Parsing ended by EOF")
 elif id == brackID:
  raise SyntaxError("SkipBody: Parsing ended by new header")
 else:
  raise SyntaxError("SkipBody: Parsing ended inside a comment")
 return None
 
def _readComments(visitor : chess.pgn.BaseVisitor, tok : ply.lex.LexToken) -> ply.lex.LexToken:
 global lexer
 comment = ''
 while tok is not None and tok.type == 'COMMENT':
  if len(comment) != 0:
   comment += '\n'
  comment += tok.value
  tok = lexer.token()
  visitor.visit_comment(comment.strip())
 return tok

def _read_variants(visitor : chess.pgn.BaseVisitor[chess.pgn.ResultT], board_stack : List[chess.Board], tok : ply.lex.LexToken, semanticError : str) -> ply.lex.LexToken:
 global lexer
 while tok is not None and tok.type == 'LPAREN':
  visitor.begin_variation()
  board = board_stack[-1].copy()
  board.pop()
  board_stack.append(board)
  tok = lexer.token()
  tok = _readComments(visitor, tok)
  tok, semanticError = _read_gameNodes(visitor, board_stack, tok, semanticError)
  _checkToken(tok, 'RPAREN')
  visitor.end_variation()
  board_stack.pop()
  tok = lexer.token()
 return tok, semanticError

def _read_gameNodes(visitor : chess.pgn.BaseVisitor[chess.pgn.ResultT], board_stack : List[chess.Board],  tok : ply.lex.LexToken, semanticError : str) -> ply.lex.LexToken:
 global lexer
 while tok is not None and tok.type == 'SANPLY':
  try:
   move =  board_stack[-1].parse_san(tok.value)
  except:
   move = chess.Move.null()
   if semanticError == '':
    semanticError = 'Parser: Improper SAN {} @ Line {}, Game {}, Pos {}'.format(tok.value, tok.lineno, lexer.gameID, tok.lexpos)
  visitor.visit_move(board_stack[-1], move)
  board_stack[-1].push(move)
  visitor.visit_board(board_stack[-1])
  tok = lexer.token()
  while tok is not None and tok.type == 'NAG':
    visitor.visit_nag(tok.value)
    tok = lexer.token()
  tok = _readComments(visitor, tok)
  if tok is not None and tok.type == 'LPAREN':
   tok, semanticError = _read_variants(visitor, board_stack, tok, semanticError)
 return tok, semanticError

def read_game(f : TextIO, Visitor : Type[chess.pgn.BaseVisitor[chess.pgn.ResultT]] = chess.pgn.GameBuilder) -> Optional[chess.pgn.ResultT]:
 """Reads a game from an open handle PGN-file in text mode.
 
By using text mode, the parser does not need to handle encodings. 
The file must be opened using the correct encoding, usually 
  
  * 'utf-8-sig': UTF-8 with optional BOM,
  * 'ascii': 7-bit ASCII,
  * 'iso-8859-1': ISO 8859/1 (Latin 1), rarely used but suggested by PGN standard.
  
Please note: *read_game* uses a private buffer.

As an extension of the PGN Standard, read_game accepts

  * null move extensions of the Standard Algebraic Notation (SAN) (0000|@@@@|Z0)
  * NAG symbols for annotations (?|??|?!|!|!!|!?)
  * NAG symbols for positions (+=|=+|+/-|+-|+--|-/+|-+|--+|=|~')
  * NAG symbols for novelty (N) and display(D)
  * Improper end of a game (see below)
    
The parser is configured by using the Visitor object leading to different *ResultT* nodes
  
  * chess.pgn.GameBuilder delivers a chess.pgn.Game object
  * chess.pgn.HeadersBuilder delivers a chess.pgn.Headers object 
  * chess.pgn.BoardBuilder delivers a chess.Board object 
  * chess.pgn.SkipVisitor delivers a boolean indicating whether a game could be successfully skipped 

The end of a game is determined by either
   
  * the game termination marker (1-0|0-1|1/2-1/2|*),
  * the end of file (EOF) occured during parsing,
  * or the start of a new header.
   
The last 2 alternatives are reported as an error, but parsing succeeds.

:param f: file handle opened in text mode 
:param Visitor: Visitor object, i.e. chess.pgn.BaseVisitor and one of the derived classes 
    
:returns: the expected *ResultT* object or ``None`` if parsing failed.
 """
 global lexer
 assert issubclass(Visitor, chess.pgn.BaseVisitor), 'Visitor ({}) must a subclass of chess.pgn.BaseVisitor'.format(type(Visitor)) 
 try:
  tok = lexer.newGame(f)
 except Exception as error:
  raise IOError('Lexer: {}'.format(error))
 if tok is None:
  return None
 visitor = Visitor()
 skipping_game = visitor.begin_game() is chess.pgn.SKIP
 headers = visitor.begin_headers()
 if headers is None:
  headers = chess.pgn.Headers()
 while tok.type == 'TAGNAME':
  tagName = tok.value
  tok = lexer.token()
  _checkToken(tok, 'TAGVALUE')
  visitor.visit_header(tagName, tok.value)
  tok = lexer.token()
 skipping_game  |= visitor.end_headers() is chess.pgn.SKIP
 tok = _readComments(visitor, tok)
 if skipping_game:
  try:
   _skipGameBody()
   lexer.endGame()
  except Exception as error:
   visitor.handle_error(error)
  visitor.end_game()
  return visitor.result()
 try:
  board_stack = [headers.board()]
  visitor.visit_board(board_stack[-1])
  tok, semanticError = _read_gameNodes(visitor,board_stack, tok, '')
  gameIsOver = board_stack[-1].is_game_over(claim_draw = False)
  if gameIsOver:
   if board_stack[-1].is_checkmate():
    if board_stack[-1].turn:
     tok.value = '0-1'
    else:
     tok.value = '1-0'
   else:
    tok.value = '1/2-1/2'
  if tok is not None and tok.type == 'TAGNAME':
   lexer.lexer.lexpos -= 1
  elif semanticError != '':
   if tok.type == 'ENDOFGAME':
    visitor.visit_result(tok.value)
   raise SyntaxError(semanticError)
  elif tok is not None:
   _checkToken(tok, 'ENDOFGAME')
   visitor.visit_result(tok.value)
   if (not gameIsOver) and headers['Result'] != tok.value:
    raise SyntaxError("Parser: game.headers['Result'] ({}) != endOfGame ({})".format(headers['Result'], tok.value))
 except Exception as error:
  try:
   visitor.handle_error(error)
  except Exception as error:
   chess.pgn.LOGGER.exception(error)
  visitor.end_game()
 lexer.endGame()
 return visitor.result()
 
def read_headers(handle : TextIO) -> Optional[chess.pgn.Headers]:
 '''Convenience function representing *read_game(handle, Visitor = chess.pgn.HeadersBuilder)*

 :param f: file handle opened in text mode 
 :returns: a *chess.pgn.Headers* object or ``None`` if parsing failed.
 '''
 return read_game(handle, Visitor = chess.pgn.HeadersBuilder)
 
def read_board(handle : TextIO) -> Optional[chess.Board]:
 '''Convenience function representing *read_game(handle, Visitor = chess.pgn.BoardBuilder)*

 :param f: file handle opened in text mode 
 :returns: a *chess.Board* object or ``None`` if parsing failed.
 '''
 return read_game(handle, Visitor = chess.pgn.BoardBuilder)

def skip_game(handle : TextIO) -> bool:
 '''Convenience function representing *read_game(handle, Visitor = chess.pgn.SkipVisitor)*

 :param f: file handle opened in text mode 
 :returns: a *bool* indicating whether parsing was successful.
 '''
 return bool(read_game(handle, Visitor = chess.pgn.SkipVisitor))

# ==================================================================
 
if __name__ == "__main__":
 import argparse
 from time import time

 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 parser = argparse.ArgumentParser(description='PGN Parser: test')
 parser.add_argument("pgnFile", help = "PGN-File (required)")
 parser.add_argument("--target",  metavar = 'target',  default = 'g', 
   choices=['g', 'game', 'h' ,'headers', 's', 'skip', 'b', 'board', 'l', 'lex'], 
   help="target of parsing (g - games, h - headers, s - skip every second game, b - board, l - lexical analysis only)")
 parser.add_argument("--encoding",  metavar = 'encoding',  default = 'u', 
   choices=['u', 'utf-8', 'i', 'iso-8859-1', 'a', 'ascii'], 
   help="target of parsing (g - games, h - headers, s - skip every second game, b - board, l - lexical analysis only)")
 parser.add_argument("-compare", action = 'store_true', default = False, help = "If True, compare with chess.pgn.read_game")
 parser.add_argument("-debug", action = 'store_true', default = False, help = "Enable Debugging")
 
 args = parser.parse_args()
 assert args.pgnFile is not None, 'pgnFile is required'
 target = args.target[0]
 encoding = args.encoding[0]
 if encoding == 'u':
  encoding = 'utf-8-sig'
 elif encoding == 'i':
  encoding = 'iso-8859-1'
 elif encoding == 'a':
  encoding = 'ascii'

 def read_board2(handle : TextIO) -> Optional[chess.Board]:
  return chess.pgn.read_game(handle, Visitor = chess.pgn.BoardBuilder)
 
 def runBoard(node, n = 100):
  if node is not None:
   for i in range(n):
    node.board()

 def runNode(readFct : Callable, f : TextIO,  gameID : int):
  try:
   node = readFct(f)
   print('node #{}, fct = {} -> {}'.format(gameID, readFct, repr(node)))
  except Exception as error:
   node = None
   print('node #{}: fct = {} failed -> {}'.format(gameID,readFct, error))
  return node

 def cmpGameNodes(chessNode, newNode):
  if isinstance(chessNode, bool) or isinstance(newNode, bool):
   context = 'Skipped game: chess.pgn: {}, new chess = {}'.format(chessNode, newNode)
   assert chessNode == newNode, context
  elif isinstance(chessNode, chess.pgn.Headers) or isinstance(newNode, chess.pgn.Headers):
   context = 'Headers: chess.pgn: {}, new chess = {}'.format(chessNode, newNode)
   assert chessNode == newNode, context
  elif isinstance(chessNode, chess.Board) or isinstance(newNode, chess.Board):
   context = 'Board: chess.pgn: {}, new chess = {}'.format(chessNode.fen(), newNode.fen())
   assert chessNode.fen() == newNode.fen(), context
  else:
   if isinstance(chessNode, chess.pgn.Game) and isinstance(newNode, chess.pgn.Game):
    context = 'Game Headers: chess.pgn: {}, new chess = {}'.format(chessNode.headers, newNode.headers)
    assert chessNode.headers == newNode.headers, context
   else:
    context = 'Move: chess.pgn: {}, new chess = {}'.format(repr(chessNode), repr(newNode))
    assert chessNode.move.uci() == newNode.move.uci(), context
   assert chessNode.comment == newNode.comment, 'Comment: chess.pgn: "{}", new chess = "{}" => {}'.format(chessNode.comment, newNode.comment, context)
   assert chessNode.nags == newNode.nags, 'Nags: chess.pgn: {}, new chess = {} => {}'.format(sorted(chessNode.nags), sorted(newNode.nags), context)
   assert len(chessNode.variations) == len(newNode.variations), 'nVariations: chess.pgn: {}, new chess = {} => {}'.format(len(chessNode.variations), len(newNode.variations), context)
   for subNode1, subNode2 in zip(chessNode.variations, newNode.variations):
    cmpGameNodes(subNode1, subNode2)

 print('File: {}'.format(os.path.abspath(args.pgnFile)))
 try:
  pgn = open(args.pgnFile, mode = 'r', encoding = encoding)
 except:
  raise IOError("Cannot open file '{}' for reading".format(args.pgnFile))
 
 if target == 'l':
  lexer = PGNLexer(bufsize = 4096, debug = args.debug, optimize = False)
  tok = lexer.newGame(pgn)
  start = time()
  lexer.dumps(tok = tok, notify = print)
  end = time()
  print('PGNLex: {:.1f} msec'.format(1000*(end-start)))
 else:
  newRead2 = None
  oldRead2 = None
  if target == 'g':
   newRead1 = read_game
   oldRead1 = chess.pgn.read_game
  elif target == 'h':
   newRead1 = read_headers
   oldRead1 = chess.pgn.read_headers
  elif target == 'b':
   newRead1 = read_board
   oldRead1 = read_board2
  elif target == 's':
   newRead1 = read_game
   oldRead1 = chess.pgn.read_game
   newRead2 = skip_game
   oldRead2 = chess.pgn.skip_game
   
  gameList = list()
  print('Parse games using new read_game ----------------------------')
  gameID = 0
  log = list()
  node = False
  start = time()
  while node is not None:
   gameID += 1
   node = runNode(newRead1, pgn, gameID)
   gameList.append(node)
   if newRead2 is not None:
    runNode(newRead2, pgn, gameID)
  end = time()
  gameID -= 1
  if newRead2 is not None:
   gameID *= 2
  log.append('new read_game: {} games read, elapsed {:.1f} msec/game'.format(gameID, 1000*(end-start)/max(gameID, 1)))
  del gameList[-1]
  if args.compare and len(gameList) > 0:
   pgn.seek(0)
   pgnGameList = list()
   print('Parse games using chess.pgn.read_game ----------------------------')
   gameID = 0
   nErrors = 0
   node = False
   start = time()
   while node is not None:
    gameID += 1
    node = runNode(oldRead1, pgn, gameID)
    pgnGameList.append(node)
    if oldRead2 is not None:
     runNode(oldRead2, pgn, gameID)
   end = time()
   gameID -= 1
   if oldRead2 is not None:
    gameID *= 2
   log.append('chess read_game: {} games read, elapsed {:.1f} msec/game'.format(gameID - 1, 1000*(end-start)/max(gameID, 1)))
   del pgnGameList[-1]
   if len(gameList) != len(pgnGameList):
    print('len(gameList) = {} != len(pgnGameList) = {}'.format(len(gameList), len(pgnGameList)))
   print('Comparing games ----------------------------')
   gameID = 0
   for node, chessNode in zip(gameList, pgnGameList):
    gameID += 1
    txt = 'game #{}: {}'.format(gameID, repr(node))
    try:
     cmpGameNodes(chessNode, node)
     print(' {} succeeded'.format(txt))
    except AssertionError as msg:
     nErrors += 1
     print(' {} failed \n -> {}'.format(txt, msg))
   print('Comparison failed for {} of {} games --------'.format(nErrors, gameID))
  print('\n'.join(log)) 
  print('completed')



'''
The use of the chessboard requires a mouse or a mousepad. A piece is moved like that: 
  
  * :kbd:`mouse-left-press`: begin move of selected piece
  * :kbd:`mouse-left-release`: end move

For training purposes, several helpers are available.

Warn of Danger
===================

The *Warn of Danger* is enabled by the ``Game/Warn of Danger`` action.
It runs as follows:

  #. assign material scores to each attack and reply for every ply
  #. create a combined  list of attacks and replies (cheapest first)
  #. create a list of total scores for every ply

Effect on other pieces like discovered check are not considered. An example:

|WarnOfDanger|

Obviously, the rook at ``b8`` is danger.

Move Options
===================

The *Move Options* are enabled by the ``Game/Show Move Options`` action.
By pressing the left mouse button at ``b8``, we see:

|MoveOptions|

It seem that ``b8-d8`` is the best move, but it is dubious ...


 .. |WarnOfDanger| image:: warnOfDanger.png
  :width: 600
  :alt: Warn of Danger
 .. |MoveOptions| image:: moveOptions.png
  :width: 600
  :alt: Move Options
'''
from typing import Optional
import os

import PyQt5.QtCore
import PyQt5.QtGui
import PyQt5.QtWidgets
import PyQt5.QtSvg

import chess, chess.pgn
import chessengine
from specialDialogs import ButtonLine
import warnOfDanger

def showStatus(board):
 print('fen = {}'.format(board.fen()))
 for row in range(8):
  for col in range(8):
   chessSquare = chess.square(col, row)
   piece = board.piece_at(chessSquare)
   if piece is not None:
    print(' square = {}, name = {}, symbol = {}'.format(
     chessSquare, chess.square_name(chessSquare), piece.symbol()))
 for n, move in enumerate(board.move_stack):
  print('{}. {}'.format(n, move.uci()))

class QBoardViewClass(PyQt5.QtWidgets.QGraphicsView):
 '''The *chessboard* is based on Qt's QGraphicsView.
 '''

 def __init__(self, parent = None) -> None:
  super(QBoardViewClass, self).__init__(parent)
  self.setFocusPolicy(PyQt5.QtCore.Qt.StrongFocus)
  self.border = 0
  self.game = Game()
  self.setMouseTracking(True)
  self.setScene(self.game)
  
 def setup(self, notifyNewGameNodeSignal : Optional[PyQt5.QtCore.pyqtSignal] = None, 
                       notifyGameNodeSelectedSignal : Optional[PyQt5.QtCore.pyqtSignal] = None, 
                      materialLabel : Optional[PyQt5.QtWidgets.QLabel] = None, 
                      squareLabel : Optional[PyQt5.QtWidgets.QLabel] = None, 
                      turnFrame : Optional[PyQt5.QtWidgets.QFrame]  = None, 
                      hintLabel : Optional[PyQt5.QtWidgets.QLabel]  = None, 
                      flipped : bool = False) -> None:
  '''Set up of the game editor
  
:param notifyNewGameNodeSignal: signal to be emitted if a move is added
:param notifyGameNodeSelectedSignal: signal to be emitted if a game node is selected
:param materialLabel: label showing the material budget
:param squareLabel: label showing the square under mouse pointer (usually in the status bar)
:param turnFrame: frame showing the color of player making the move
:param hintLabel: label showing the engine's hints and scores (usually in the status bar)
:param flipped: if True, the square *a1* is on top
  '''
  self.notifyGameNodeSelectedSignal = notifyGameNodeSelectedSignal 
  self.game.setup(notifyNewGameNodeSignal = notifyNewGameNodeSignal,  materialLabel = materialLabel, squareLabel = squareLabel, 
   hintLabel = hintLabel, turnFrame =turnFrame,  flipped = flipped)
  self._configureScene()
  self.game.draw_board()

 def setHint(self, enableHint : bool = False, enableScore : bool = False, 
                         engine : Optional[chessengine.ChessEngine] = None) -> None:
  '''Controls the usage of the hint label

:param enableHint: If True, the *hintLabel* shows the *engine* hint
:param enableScore: If True, the *hintLabel* shows the *engine* scores
:param engine: engine to be used (required, if *enableHint or enableScore*
  '''
  self.game.setHint(enableHint, enableScore,  engine = engine)

 def setFlipped(self, enable : bool) -> None:
  '''Controls the board orientation

:param enable: if True, the square *a1* is on top
  '''
  self.game.setFlipped(enable)

 def setDrawOptions(self, enable : bool) -> None:
  '''Controls the draw options. Draw options are shown when :kbd:`mouse-left-press` on a piece

:param enable: if True, the draw options are enabled
  '''
  self.game.setDrawOptions(enable)
  
 def setWarnOfDanger(self, enable : bool) -> None:
  '''Controls the warn of danger, i.e. shows attacked pieces

:param enable: if True, the warn of danger is enabled
  '''
  self.game.setWarnOfDanger(enable)
  
 def _fen(self) -> str:
  return self.game.fen()
 
 def set_fen(self, fen : str) -> None:
  '''Configures the *chessboard*

:param fen: position in `Forsyth-Edwards` Notation
  .. _Forsyth-Edwards: https://github.com/fsmosca/PGN-Standard
  '''
  self.game.set_fen(fen)
 
 def setGameNode(self, gameNode : chess.pgn.GameNode) -> None:
  '''Sets the game node displayed by the *chessBoard* 
  
:param gameNode : game node to be displayed
  '''
  self.game.setGameNode(gameNode)
  
 def nextMove(self) -> None:
  '''Go 1 move forward, if possible
  '''
  return self.game.nextMove()
 
 def previousMove(self) -> None:
  '''Go 1 move backward, if possible
  '''
  return self.game.previousMove()

 def _configureScene(self):
  self.fitInView(PyQt5.QtCore.QRectF(0, 0, 8 * Game.pieceSize, 8 * Game.pieceSize), mode = PyQt5.QtCore.Qt.KeepAspectRatio)
  # self.setScene(self.game)
  
 def resizeEvent(self, ev) -> None:
  # print('oldSize = {}, size = {}'.format(ev.oldSize(), ev.size()))
  PyQt5.QtWidgets.QGraphicsView.resizeEvent(self, ev)
  self._configureScene()
  if self.game.pressedPiece is None:
   self.game.draw_board()
   
 @PyQt5.QtCore.pyqtSlot(PyQt5.QtGui.QKeyEvent)
 def keyPressEvent(self, ev):
  if ev.key() == PyQt5.QtCore.Qt.Key_Up:
   newGameNode = self.game.gameNode.parent
  else:
   newGameNode = self.game.gameNode.next()
  if newGameNode is not None:
   self.notifyGameNodeSelectedSignal.emit(newGameNode)

class Piece(PyQt5.QtSvg.QGraphicsSvgItem):
 '''Internal class
 '''
 piecesDirectory = os.path.join(os.path.dirname(os.path.abspath(__file__)), ':/pieces/')
 colorChars = ('black', 'white')

 def __init__(self, symbolName : str, size : int, parent = None) -> None:
  '''
  symbol values (according to PGN spec)
   p,P = white pawn, black ~
   n,N = white knight, black ~ 
   b,B = white bishop, black ~ 
   r,R = white rook, black ~ 
   q,Q = white queen, black ~
   k,K = white king, black ~
  '''
  self.llPiece = chess.Piece.from_symbol(symbolName)
  fileDirectory = os.path.dirname(os.path.abspath(__file__))
  svgPath = os.path.join(fileDirectory , 'pieces', 
    '{}-{}.svg'.format(self.colorChars[self.llPiece.color], chess.piece_name(self.llPiece.piece_type)))
  super(Piece, self).__init__(svgPath, parent)
  self.setFlags(PyQt5.QtWidgets.QGraphicsItem.ItemIsMovable | PyQt5.QtWidgets.QGraphicsItem.ItemIsSelectable)
  bwidth = self.boundingRect().width()
  self.setScale(size/bwidth)
  # self.setAcceptHoverEvents(True)
  self.setCacheMode(PyQt5.QtWidgets.QGraphicsItem.NoCache)
  self.setZValue(1)

class Game(PyQt5.QtWidgets.QGraphicsScene):
 '''Internal class
 '''
 whiteSquareBrush = PyQt5.QtGui.QBrush( PyQt5.QtCore.Qt.white, PyQt5.QtCore.Qt.SolidPattern)
 blackSquareBrush = PyQt5.QtGui.QBrush( PyQt5.QtCore.Qt.gray, PyQt5.QtCore.Qt.SolidPattern)
 goodSquareBrush = PyQt5.QtGui.QBrush( PyQt5.QtCore.Qt.green, PyQt5.QtCore.Qt.SolidPattern)
 intermediateSquareBrush = PyQt5.QtGui.QBrush( PyQt5.QtCore.Qt.yellow, PyQt5.QtCore.Qt.SolidPattern)
 badSquareBrush = PyQt5.QtGui.QBrush( PyQt5.QtCore.Qt.red, PyQt5.QtCore.Qt.SolidPattern)
 requestHint = PyQt5.QtCore.pyqtSignal(str)
 pieceSize = 45
 leipzigEncodeDict = {
   'P' : 'p', 'N' : 'n', 'B' : 'b', 'R' : 'r', 'Q' : 'q', 
   'p' : 'o', 'n' : 'm', 'b' : 'v', 'r' : 't', 'q' : 'w' 
  }
 chessFigures = { '♕' :  chess.QUEEN, '♖' :  chess.ROOK,  '♗' :  chess.BISHOP,  '♘' :  chess.KNIGHT}
 chessFiguresDescription =  { '♕' : 'Queen', '♖' : 'Rook', '♗' :  'Bishop', '♘' :  'Knight'}
 encodePiece = lambda piece_type, color : Game.leipzigEncodeDict[chess.Piece(piece_type, color).symbol()]
 
 def __init__(self, parent = None) -> None:
  super(Game, self).__init__(parent)
  
  self.timer = PyQt5.QtCore.QTimer()
  self.timer.timeout.connect(self.flushHint)
  self.requestHint.connect(self.hintRequested)
  self.hintDelay = 100

  self.promotePawn = ButtonLine(self.chessFigures, hintDict = self.chessFiguresDescription, title = 'Promotion', pointSize = 30)
  self.gameNode = chess.pgn.Game()

  self.boardElementGroup = None
  self.drawOptionsGroup = list()
  self.boardPieces = list()
  self.flipped = False
  self.engine = None
  self.hint = None
  self.drawOptions = False
  self.warnOfDanger = False
  self.materialLabel = None
  self.squareLabel = None
  self.turnFrame = None
  self.hintLabel = None
  
 def setup(self, notifyNewGameNodeSignal : Optional[PyQt5.QtCore.pyqtSignal] = None, 
                      materialLabel : Optional[PyQt5.QtWidgets.QLabel] = None, 
                      squareLabel : Optional[PyQt5.QtWidgets.QLabel] = None, 
                      turnFrame : Optional[PyQt5.QtWidgets.QFrame]  = None, 
                      hintLabel : Optional[PyQt5.QtWidgets.QLabel]  = None, 
                      flipped : bool = False, 
                      gameNode : Optional[chess.pgn.GameNode]= None) -> None:
  self.notifyNewGameNodeSignal = notifyNewGameNodeSignal
  self.materialLabel = materialLabel
  self.squareLabel = squareLabel
  self.turnFrame = turnFrame
  self.hintLabel = hintLabel
  if gameNode is not None:
   self.gameNode = gameNode
  self.boardElementGroup = None
  self.boardPieces = list()
  self.flipped = flipped
  board = self.gameNode.board()
  self.showMaterial(board)
  self.showTurn(board)
  self.pressedPiece = None
  
 def boardRect(self) -> PyQt5.QtCore.QRectF:
  return PyQt5.QtCore.QRectF(0, 0, 8 * self.pieceSize, 8 * self.pieceSize)

 def setDrawOptions(self, enable : bool) -> None:
  self.drawOptions = enable
  
 def setWarnOfDanger(self, enable : bool) -> None:
  if enable != self.warnOfDanger:
   self.warnOfDanger = enable
   if enable:
    self.draw_warnOfDanger()
   else:
    self.remove_warnOfDanger()
  
 def fen(self):
  return self.gameNode.board().fen()
  
 def setGameNode(self, game : bool) -> None:
  if game is None:
   return
  self.gameNode = game
  self.draw_board()
  board = self.gameNode.board()
  self.showMaterial(board)
  self.isGameOver = False
  self.showTurn(board)
  if self.hint:
   self.requestHint.emit(self.fen())
  self.draw_warnOfDanger()

 # -------------------------------------------------------

 def nextMove(self) -> None:
  newGameNode = self.gameNode.next()
  if newGameNode is None:
   return self.gameNode
  self.gameNode = newGameNode
  self.draw_board()
  board = self.gameNode.board()
  self.showTurn(board)
  if self.hint:
    self.requestHint.emit(self.fen())
  self.showMaterial(board)
  self.showTurn(board)
  self.draw_warnOfDanger()
  return newGameNode
 
 def previousMove(self) -> None:
  newGameNode = self.gameNode.parent
  if newGameNode is None:
   return self.gameNode
  self.gameNode = newGameNode
  self.draw_board()
  board = self.gameNode.board()
  self.showTurn(board)
  if self.hint:
   self.requestHint.emit(self.fen())
  self.showMaterial(board)
  self.showTurn(board)
  self.draw_warnOfDanger()
  return newGameNode

 # -------------------------------------------------------

 def showTurn(self, board : chess.Board) -> None:
  if self.turnFrame is None:
   return
  self.gameIsOver = board.is_game_over()
  game = self.gameNode.game()
  if self.gameIsOver or (game.headers['Result'] != '*' and self.gameNode.is_end()):
   if self.gameIsOver:
    game.headers['Result'] = board.result()
    if board.is_checkmate():
     self.turnFrame.setStyleSheet('background-color: red')
    else:
     self.turnFrame.setStyleSheet('background-color: yellow')
   elif game.headers['Result'] == '1/2-1/2':
    self.turnFrame.setStyleSheet('background-color: yellow')
   else:
    self.turnFrame.setStyleSheet('background-color: red')
   self.gameIsOver = True
  elif board.turn:
   self.turnFrame.setStyleSheet('background-color: white')
  else:
   self.turnFrame.setStyleSheet('background-color: black')

 def showMaterial(self, board : chess.Board) -> None:
  if self.materialLabel is None:
   return
  pieceMap = board.piece_map()
  whiteCountList = 7 * [0]
  blackCountList = 7 * [0]
  for piece in list(pieceMap.values()):
   if piece.color == chess.WHITE:
    whiteCountList[piece.piece_type] += 1
   else:
    blackCountList[piece.piece_type] += 1
  piece_text =''
  for piece_type in range(5, 0,-1):
   delta = whiteCountList[piece_type] - blackCountList[piece_type]
   if delta == 0:
    continue
   if delta > 0:
    piece_text += delta * Game.encodePiece(piece_type, chess.WHITE)
   else:
    piece_text += abs(delta) * Game.encodePiece(piece_type, chess.BLACK)
  self.materialLabel.setText(piece_text)

 # -------------------------------------------------------

 def draw_board(self, flipped : Optional[bool] = None) -> None:
  self.clear()
  elementWidth = self.pieceSize
  for row in range(8):
   for col in range(8):
    rect = PyQt5.QtWidgets.QGraphicsRectItem( row * elementWidth, col * elementWidth, elementWidth, elementWidth)
    if (row + self.flipped) % 2 == col % 2:
     rect.setBrush( self.whiteSquareBrush )
    else:
     rect.setBrush( self.blackSquareBrush )
    rect.setCacheMode(PyQt5.QtWidgets.QGraphicsItem.NoCache)
    rect.setZValue(0)
    self.addItem(rect)
  if flipped is not None:
   self.flipped = flipped
  self.draw_pieces()
  
 def draw_pieces(self) -> None:
  symbolWidth = self.pieceSize
  for row in range(8):
   for col in range(8):
    chessSquare = chess.square(col, row)
    piece = self.gameNode.board().piece_at(chessSquare)
    if piece is not None:
     # print('draw_pieces: square = {}, name = {}'.format(chessSquare, chess.square_name(chessSquare)))
     # print('draw_pieces: piece = {}, scenePos = {}'.format(piece.symbol(), self.getScenePos(chessSquare)))
     bPiece = Piece(piece.symbol(), size = symbolWidth)
     bPiece.setPos(self.getScenePos(chessSquare))
     bPiece.setScale(symbolWidth/bPiece.boundingRect().width())
     self.addItem(bPiece)
  self.update()

 def draw_drawOptions(self) -> None:
  if len(self.legal_targets) == 0 or not self.drawOptions:
   return 
  elementSize = PyQt5.QtCore.QSizeF(self.pieceSize, self.pieceSize)
  board = self.gameNode.board()
  self.drawOptionsGroup = len(self.legal_targets)*[None]
  for n, chessSquare in enumerate(self.legal_targets):
   self.drawOptionsGroup[n] = PyQt5.QtWidgets.QGraphicsRectItem(
              PyQt5.QtCore.QRectF(self.getScenePos(chessSquare), elementSize))
   self.drawOptionsGroup[n].setCacheMode(PyQt5.QtWidgets.QGraphicsItem.NoCache)
   if board.is_attacked_by(not board.turn, chessSquare):
    self.drawOptionsGroup[n].setBrush(self.badSquareBrush)
   else:
    self.drawOptionsGroup[n].setBrush(self.goodSquareBrush)
   self.drawOptionsGroup[n].setZValue(0.8)
   self.addItem(self.drawOptionsGroup[n])
  self.update()

 def remove_drawOptions(self) -> None:
  try:
   for rect in self.drawOptionsGroup:
    self.removeItem(rect)
   self.update()
  except:
   pass
  self.drawOptionsGroup = list()
   
 def draw_warnOfDanger(self) -> None:
  if not self.warnOfDanger:
   return 
  elementSize = PyQt5.QtCore.QSizeF(self.pieceSize, self.pieceSize)
  square2ScoreDict = warnOfDanger.warnOfDanger(self.gameNode)
  if square2ScoreDict is None or len(square2ScoreDict) == 0:
   return
  self.warnOfDangerGroup = len(square2ScoreDict)*[None]
  for n, chessSquare in enumerate(square2ScoreDict):
   score = square2ScoreDict[chessSquare]
   self.warnOfDangerGroup[n] = PyQt5.QtWidgets.QGraphicsRectItem(
              PyQt5.QtCore.QRectF(self.getScenePos(chessSquare), elementSize))
   self.warnOfDangerGroup[n].setCacheMode(PyQt5.QtWidgets.QGraphicsItem.NoCache)
   if score > 0:
    self.warnOfDangerGroup[n].setBrush(self.badSquareBrush)
   elif score == 0:
    self.warnOfDangerGroup[n].setBrush(self.intermediateSquareBrush)
   else:
    self.warnOfDangerGroup[n].setBrush(self.goodSquareBrush)
   self.warnOfDangerGroup[n].setZValue(0.4)
   self.addItem(self.warnOfDangerGroup[n])
  self.update()

 def remove_warnOfDanger(self) -> None:
  try:
   for rect in self.warnOfDangerGroup:
    self.removeItem(rect)
  except:
   pass
  self.warnOfDangerGroup = list()
   
 def getScenePos(self, chessSquare : int) -> PyQt5.QtCore.QPointF:
  if chessSquare < 0 or chessSquare > 63:
   return None 
  col = chess.square_file(chessSquare)
  if self.flipped:
   row = chess.square_rank(chessSquare)
  else:
   row = 7 - chess.square_rank(chessSquare)
  return PyQt5.QtCore.QPointF(col * self.pieceSize, row * self.pieceSize)

 # -------------------------------------------------------
   
 def getChessSquareAt(self, scenePos : PyQt5.QtCore.QPointF) -> int:
  iPos = scenePos/self.pieceSize
  if self.flipped:
   id = chess.square(int(iPos.x()), int(iPos.y()))
  else:
   id = chess.square(int(iPos.x()), 7 - int(iPos.y()))
  if id < 0 or id > 63:
   return None 
  return id

 # -------------------------------------------------------
 
 def setHint(self, enableHint : bool = False, enableScore : bool = False, 
                         engine : Optional[chessengine.ChessEngine] = None) -> None:
  self.engine = engine
  self.hint = (self.engine is not None and enableHint)
  self.score = (self.engine is not None and enableScore)
  if self.hint or self.score:
   self.hintFen = None
   self.engine.bestMoveScoreSignal.connect(self.bestMoveScoreAvailable)
   self.requestHint.emit(self.fen())
  else:
   self.hintLabel.setText('-/-')

 def setFlipped(self, enable : bool) -> None:
  self.flipped = enable
  self.setGameNode(self.gameNode)
  
 def hintRequested(self, fen : str) -> None:
  if not (self.hint or self.score):
   return
  self.timer.stop()
  self.timer.start(self.hintDelay+1)
 
 def flushHint(self) -> None:
  fen = self.fen()
  if self.hintFen != fen and self.engine.uciNewGame(fen):
   self.engine.startPlay()
   self.hintFen = fen

 @PyQt5.QtCore.pyqtSlot(chess.Move, str)
 def bestMoveScoreAvailable(self, move, score):
  if self.hintLabel is not None:
   board = self.gameNode.board()
   if self.hint:
    try:
     moveString = board.lan(move)
    except:
     moveString = '?{}'.format(move.uci())
   else:
    moveString = '-'
   if self.score:
    self.hintLabel.setText('{}/{}'.format(moveString, score))
   else:
    self.hintLabel.setText('{}/-'.format(moveString))

 # -------------------------------------------------------

 def mouseMoveEvent(self, qGraphicsSceneMouseEvent : PyQt5.QtWidgets.QGraphicsSceneMouseEvent) -> None:
  if self.squareLabel is not None:
   square = self.getChessSquareAt(qGraphicsSceneMouseEvent.scenePos())
   if square is not None:
    self.squareLabel.setText(chess.square_name(square))
   else:
    self.squareLabel.setText('--')
  PyQt5.QtWidgets.QGraphicsScene.mouseMoveEvent(self, qGraphicsSceneMouseEvent)

 def mousePressEvent(self, qGraphicsSceneMouseEvent : PyQt5.QtWidgets.QGraphicsSceneMouseEvent) -> None:
  self.pressedPiece = None
  self.remove_warnOfDanger()
  self.pressedSquareId = self.getChessSquareAt(qGraphicsSceneMouseEvent.scenePos())
  if self.pressedSquareId is not None:
   self.pressedPiece = self.itemAt(self.getScenePos(self.pressedSquareId), PyQt5.QtGui.QTransform())
   if isinstance(self.pressedPiece,PyQt5.QtWidgets.QGraphicsRectItem):
    self.pressedPiece = None # empty field hitten
  board = self.gameNode.board()
  if not self.gameIsOver:
   self.pressedSquareId = self.getChessSquareAt(qGraphicsSceneMouseEvent.scenePos())
   if self.pressedPiece is not None:
    self.pressedPiece = self.itemAt(self.getScenePos(self.pressedSquareId), PyQt5.QtGui.QTransform())
    self.legal_targets = list()
    for move in list(board.legal_moves):
     if move.from_square == self.pressedSquareId:
      self.legal_targets.append(move.to_square)
    self.attackedPieceDict = dict()
    for pieceID in board.attacks(self.pressedSquareId): 
     actPiece = self.itemAt(self.getScenePos(pieceID), PyQt5.QtGui.QTransform())
     if not isinstance(actPiece,PyQt5.QtWidgets.QGraphicsRectItem):
      self.attackedPieceDict[pieceID] = actPiece
    self.draw_drawOptions()
  PyQt5.QtWidgets.QGraphicsScene.mousePressEvent(self, qGraphicsSceneMouseEvent)

 def mouseReleaseEvent(self, qGraphicsSceneMouseEvent : PyQt5.QtWidgets.QGraphicsSceneMouseEvent) -> None:
  PyQt5.QtWidgets.QGraphicsScene.mouseReleaseEvent(self, qGraphicsSceneMouseEvent)
  if self.pressedPiece is None or self.gameIsOver:
   if self.pressedPiece is not None:
    self.pressedPiece.setPos(self.getScenePos(self.pressedSquareId))
   return
  self.remove_drawOptions()
  releasedSquareId = self.getChessSquareAt(qGraphicsSceneMouseEvent.scenePos())
  if releasedSquareId in self.legal_targets:
   if releasedSquareId in self.attackedPieceDict:
    self.removeItem(self.attackedPieceDict[releasedSquareId])
  else:
   releasedSquareId = self.pressedSquareId
  if releasedSquareId != self.pressedSquareId:
   oldBoard = self.gameNode.board()
   if self.pressedPiece.llPiece.piece_type == chess.PAWN and \
     ((self.pressedPiece.llPiece.color == chess.WHITE and releasedSquareId > chess.H7) \
     or (self.pressedPiece.llPiece.color == chess.BLACK and releasedSquareId < chess.A2)):
    self.promotePawn.move(qGraphicsSceneMouseEvent.screenPos())
    promotedPieceType = self.promotePawn.exec()
    move = chess.Move(self.pressedSquareId, releasedSquareId, promotion = promotedPieceType)
    self.removeItem(self.pressedPiece)
    promotedPiece = chess.Piece(promotedPieceType, self.pressedPiece.llPiece.color)
    self.addItem(Piece(promotedPiece.symbol(), self.pieceSize))
    self.update()
   else:
    move = chess.Move(self.pressedSquareId, releasedSquareId)
    if oldBoard.is_castling(move): 
     isWhite = (self.pressedPiece.llPiece.color == chess.WHITE)
     if (isWhite and self.pressedSquareId != chess.E1) \
       or ((not isWhite) and self.pressedSquareId != chess.E8):
      raise ValueError('UIE: Unexpected castling move')
     if releasedSquareId > self.pressedSquareId:
      if isWhite:
       rookPos = self.getScenePos(chess.H1)
       rookTargetPos = self.getScenePos(chess.F1)
      else:
       rookPos = self.getScenePos(chess.H8)
       rookTargetPos = self.getScenePos(chess.F8)
     else:
      if isWhite:
       rookPos = self.getScenePos(chess.A1)
       rookTargetPos = self.getScenePos(chess.D1)
      else:
       rookPos = self.getScenePos(chess.A8)
       rookTargetPos = self.getScenePos(chess.D8)
     rookPiece = self.itemAt(rookPos, PyQt5.QtGui.QTransform())
     if isinstance(rookPiece,PyQt5.QtWidgets.QGraphicsRectItem):
      raise ValueError('UIE: QGraphicsRectItem found')
      rookPiece.setPos(self.getScenePos(releasedSquareId))
     rookPiece.setPos(rookTargetPos )
    self.pressedPiece.setPos(self.getScenePos(releasedSquareId))
   self.gameNode = self.gameNode.add_variation(move)
   board = self.gameNode.board()
   # showStatus(board)
   self.showMaterial(board)
   self.showTurn(board)
   if self.notifyNewGameNodeSignal is not None:
    self.notifyNewGameNodeSignal.emit(self.gameNode)
   if self.hint:
    self.requestHint.emit(self.fen())
  else:
   self.pressedPiece.setPos(self.getScenePos(releasedSquareId))
  self.pressedPiece = None
  self.draw_warnOfDanger()




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Users\Reinh\OneDrive\Dokumente\python\MzChess\MzChess\chessMainWindow.ui'
#
# Created by: PyQt5 UI code generator 5.15.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(973, 753)
        self.centralWidget = QtWidgets.QWidget(MainWindow)
        self.centralWidget.setObjectName("centralWidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralWidget)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.splitter = QtWidgets.QSplitter(self.centralWidget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.boardGraphicsView = QBoardViewClass(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(5)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.boardGraphicsView.sizePolicy().hasHeightForWidth())
        self.boardGraphicsView.setSizePolicy(sizePolicy)
        self.boardGraphicsView.setMinimumSize(QtCore.QSize(100, 100))
        self.boardGraphicsView.setObjectName("boardGraphicsView")
        self.verticalLayout.addWidget(self.boardGraphicsView)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.materialLabel = QtWidgets.QLabel(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.materialLabel.sizePolicy().hasHeightForWidth())
        self.materialLabel.setSizePolicy(sizePolicy)
        self.materialLabel.setMinimumSize(QtCore.QSize(0, 30))
        self.materialLabel.setMaximumSize(QtCore.QSize(16777215, 30))
        font = QtGui.QFont()
        font.setFamily("Chess Leipzig")
        font.setPointSize(20)
        self.materialLabel.setFont(font)
        self.materialLabel.setText("")
        self.materialLabel.setObjectName("materialLabel")
        self.horizontalLayout.addWidget(self.materialLabel)
        self.turnFrame = QtWidgets.QFrame(self.layoutWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.turnFrame.sizePolicy().hasHeightForWidth())
        self.turnFrame.setSizePolicy(sizePolicy)
        self.turnFrame.setMinimumSize(QtCore.QSize(30, 30))
        self.turnFrame.setAutoFillBackground(True)
        self.turnFrame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.turnFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.turnFrame.setObjectName("turnFrame")
        self.horizontalLayout.addWidget(self.turnFrame)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.tabWidget = QtWidgets.QTabWidget(self.splitter)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setMinimumSize(QtCore.QSize(100, 0))
        self.tabWidget.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.tabWidget.setObjectName("tabWidget")
        self.tabGame = QtWidgets.QWidget()
        self.tabGame.setObjectName("tabGame")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tabGame)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.gameTreeViewWidget = GameTreeView(self.tabGame)
        self.gameTreeViewWidget.setObjectName("gameTreeViewWidget")
        self.gameTreeViewWidget.headerItem().setText(0, "1")
        self.verticalLayout_3.addWidget(self.gameTreeViewWidget)
        self.tabWidget.addTab(self.tabGame, "")
        self.tabHeader = QtWidgets.QWidget()
        self.tabHeader.setObjectName("tabHeader")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.tabHeader)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.gameHeaderTableView = GameHeaderView(self.tabHeader)
        self.gameHeaderTableView.setEditTriggers(QtWidgets.QAbstractItemView.EditKeyPressed)
        self.gameHeaderTableView.setColumnCount(2)
        self.gameHeaderTableView.setObjectName("gameHeaderTableView")
        self.gameHeaderTableView.setRowCount(0)
        self.gameHeaderTableView.horizontalHeader().setVisible(False)
        self.gameHeaderTableView.horizontalHeader().setStretchLastSection(True)
        self.verticalLayout_4.addWidget(self.gameHeaderTableView)
        self.tabWidget.addTab(self.tabHeader, "")
        self.tabDatabase = QtWidgets.QWidget()
        self.tabDatabase.setObjectName("tabDatabase")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.tabDatabase)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.gameListTableView = GameListTableView(self.tabDatabase)
        font = QtGui.QFont()
        font.setFamily("Courier")
        font.setPointSize(10)
        self.gameListTableView.setFont(font)
        self.gameListTableView.setObjectName("gameListTableView")
        self.horizontalLayout_3.addWidget(self.gameListTableView)
        self.tabWidget.addTab(self.tabDatabase, "")
        self.tabScorePlot = QtWidgets.QWidget()
        self.tabScorePlot.setObjectName("tabScorePlot")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.tabScorePlot)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.scorePlotGraphicsView = ScorePlot(self.tabScorePlot)
        self.scorePlotGraphicsView.setObjectName("scorePlotGraphicsView")
        self.horizontalLayout_2.addWidget(self.scorePlotGraphicsView)
        self.tabWidget.addTab(self.tabScorePlot, "")
        self.tabLog = QtWidgets.QWidget()
        self.tabLog.setObjectName("tabLog")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.tabLog)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.uciTextEdit = QUCIEdit(self.tabLog)
        self.uciTextEdit.setObjectName("uciTextEdit")
        self.horizontalLayout_4.addWidget(self.uciTextEdit)
        self.tabWidget.addTab(self.tabLog, "")
        self.verticalLayout_2.addWidget(self.splitter)
        MainWindow.setCentralWidget(self.centralWidget)
        self.menuBar = QtWidgets.QMenuBar(MainWindow)
        self.menuBar.setGeometry(QtCore.QRect(0, 0, 973, 21))
        self.menuBar.setObjectName("menuBar")
        self.menuFile = QtWidgets.QMenu(self.menuBar)
        self.menuFile.setObjectName("menuFile")
        self.menuRecent_PGN = QtWidgets.QMenu(self.menuFile)
        self.menuRecent_PGN.setObjectName("menuRecent_PGN")
        self.menuEncoding = QtWidgets.QMenu(self.menuFile)
        self.menuEncoding.setObjectName("menuEncoding")
        self.menuEdit = QtWidgets.QMenu(self.menuBar)
        self.menuEdit.setObjectName("menuEdit")
        self.menuGame = QtWidgets.QMenu(self.menuBar)
        self.menuGame.setObjectName("menuGame")
        self.menuEnd_Game = QtWidgets.QMenu(self.menuGame)
        self.menuEnd_Game.setObjectName("menuEnd_Game")
        self.menuEngine = QtWidgets.QMenu(self.menuBar)
        self.menuEngine.setObjectName("menuEngine")
        self.menuSelect_Engine = QtWidgets.QMenu(self.menuEngine)
        self.menuSelect_Engine.setObjectName("menuSelect_Engine")
        self.menuSearch_Depth = QtWidgets.QMenu(self.menuEngine)
        self.menuSearch_Depth.setObjectName("menuSearch_Depth")
        self.menuBlunder_Limit = QtWidgets.QMenu(self.menuEngine)
        self.menuBlunder_Limit.setObjectName("menuBlunder_Limit")
        self.menuAnnotate_Variants = QtWidgets.QMenu(self.menuEngine)
        self.menuAnnotate_Variants.setObjectName("menuAnnotate_Variants")
        self.menuNumber_of_Annotations = QtWidgets.QMenu(self.menuEngine)
        self.menuNumber_of_Annotations.setObjectName("menuNumber_of_Annotations")
        self.menuDatabase = QtWidgets.QMenu(self.menuBar)
        self.menuDatabase.setObjectName("menuDatabase")
        self.menuHelp = QtWidgets.QMenu(self.menuBar)
        self.menuHelp.setObjectName("menuHelp")
        MainWindow.setMenuBar(self.menuBar)
        self.actionCopy_FEN = QtWidgets.QAction(MainWindow)
        self.actionCopy_FEN.setObjectName("actionCopy_FEN")
        self.actionPaste_FEN = QtWidgets.QAction(MainWindow)
        self.actionPaste_FEN.setObjectName("actionPaste_FEN")
        self.actionCopy_PGN = QtWidgets.QAction(MainWindow)
        self.actionCopy_PGN.setObjectName("actionCopy_PGN")
        self.actionPaste_PGN = QtWidgets.QAction(MainWindow)
        self.actionPaste_PGN.setObjectName("actionPaste_PGN")
        self.actionNew_Game = QtWidgets.QAction(MainWindow)
        self.actionNew_Game.setObjectName("actionNew_Game")
        self.actionWhite_wins = QtWidgets.QAction(MainWindow)
        self.actionWhite_wins.setObjectName("actionWhite_wins")
        self.action0_1 = QtWidgets.QAction(MainWindow)
        self.action0_1.setObjectName("action0_1")
        self.action1_2_1_2 = QtWidgets.QAction(MainWindow)
        self.action1_2_1_2.setObjectName("action1_2_1_2")
        self.actionShow_Options = QtWidgets.QAction(MainWindow)
        self.actionShow_Options.setCheckable(True)
        self.actionShow_Options.setObjectName("actionShow_Options")
        self.action5_Moves = QtWidgets.QAction(MainWindow)
        self.action5_Moves.setCheckable(True)
        self.action5_Moves.setObjectName("action5_Moves")
        self.action10_Moves = QtWidgets.QAction(MainWindow)
        self.action10_Moves.setCheckable(True)
        self.action10_Moves.setObjectName("action10_Moves")
        self.action15_Moves = QtWidgets.QAction(MainWindow)
        self.action15_Moves.setCheckable(True)
        self.action15_Moves.setChecked(False)
        self.action15_Moves.setObjectName("action15_Moves")
        self.action20_Moves = QtWidgets.QAction(MainWindow)
        self.action20_Moves.setCheckable(True)
        self.action20_Moves.setObjectName("action20_Moves")
        self.action25_Moves = QtWidgets.QAction(MainWindow)
        self.action25_Moves.setCheckable(True)
        self.action25_Moves.setObjectName("action25_Moves")
        self.actionShow_Hints = QtWidgets.QAction(MainWindow)
        self.actionShow_Hints.setCheckable(True)
        self.actionShow_Hints.setObjectName("actionShow_Hints")
        self.actionNext_Move = QtWidgets.QAction(MainWindow)
        self.actionNext_Move.setShortcutContext(QtCore.Qt.WidgetShortcut)
        self.actionNext_Move.setObjectName("actionNext_Move")
        self.actionPrevious_Move = QtWidgets.QAction(MainWindow)
        self.actionPrevious_Move.setShortcutContext(QtCore.Qt.WidgetShortcut)
        self.actionPrevious_Move.setObjectName("actionPrevious_Move")
        self.actionOpen_PGN = QtWidgets.QAction(MainWindow)
        self.actionOpen_PGN.setObjectName("actionOpen_PGN")
        self.actionSave_PGN = QtWidgets.QAction(MainWindow)
        self.actionSave_PGN.setObjectName("actionSave_PGN")
        self.actionExit = QtWidgets.QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.actionbla = QtWidgets.QAction(MainWindow)
        self.actionbla.setObjectName("actionbla")
        self.actionAdd_To_DB = QtWidgets.QAction(MainWindow)
        self.actionAdd_To_DB.setObjectName("actionAdd_To_DB")
        self.actionAppend_to_DB = QtWidgets.QAction(MainWindow)
        self.actionAppend_to_DB.setCheckable(True)
        self.actionAppend_to_DB.setObjectName("actionAppend_to_DB")
        self.actionSave_Game = QtWidgets.QAction(MainWindow)
        self.actionSave_Game.setObjectName("actionSave_Game")
        self.actionClose_PGN = QtWidgets.QAction(MainWindow)
        self.actionClose_PGN.setObjectName("actionClose_PGN")
        self.actionAdd_Game = QtWidgets.QAction(MainWindow)
        self.actionAdd_Game.setObjectName("actionAdd_Game")
        self.actionRemove_Games = QtWidgets.QAction(MainWindow)
        self.actionRemove_Games.setObjectName("actionRemove_Games")
        self.actionNew_PGN = QtWidgets.QAction(MainWindow)
        self.actionNew_PGN.setObjectName("actionNew_PGN")
        self.actionFlip_Board = QtWidgets.QAction(MainWindow)
        self.actionFlip_Board.setCheckable(True)
        self.actionFlip_Board.setObjectName("actionFlip_Board")
        self.actionNext_Variant = QtWidgets.QAction(MainWindow)
        self.actionNext_Variant.setObjectName("actionNext_Variant")
        self.actionPrevious_Variant = QtWidgets.QAction(MainWindow)
        self.actionPrevious_Variant.setObjectName("actionPrevious_Variant")
        self.actionShow_Scores = QtWidgets.QAction(MainWindow)
        self.actionShow_Scores.setCheckable(True)
        self.actionShow_Scores.setObjectName("actionShow_Scores")
        self.actionWarn_of_Danger = QtWidgets.QAction(MainWindow)
        self.actionWarn_of_Danger.setCheckable(True)
        self.actionWarn_of_Danger.setObjectName("actionWarn_of_Danger")
        self.actionUndo_Last_Move = QtWidgets.QAction(MainWindow)
        self.actionUndo_Last_Move.setObjectName("actionUndo_Last_Move")
        self.actionAnnotate_Last_Move = QtWidgets.QAction(MainWindow)
        self.actionAnnotate_Last_Move.setObjectName("actionAnnotate_Last_Move")
        self.actionBlNone = QtWidgets.QAction(MainWindow)
        self.actionBlNone.setCheckable(True)
        self.actionBlNone.setChecked(False)
        self.actionBlNone.setObjectName("actionBlNone")
        self.action0_5_pawns = QtWidgets.QAction(MainWindow)
        self.action0_5_pawns.setCheckable(True)
        self.action0_5_pawns.setObjectName("action0_5_pawns")
        self.action1_pawn = QtWidgets.QAction(MainWindow)
        self.action1_pawn.setCheckable(True)
        self.action1_pawn.setObjectName("action1_pawn")
        self.action1_5_pawns = QtWidgets.QAction(MainWindow)
        self.action1_5_pawns.setCheckable(True)
        self.action1_5_pawns.setObjectName("action1_5_pawns")
        self.action2_0_pawns = QtWidgets.QAction(MainWindow)
        self.action2_0_pawns.setCheckable(True)
        self.action2_0_pawns.setObjectName("action2_0_pawns")
        self.action2_5_pawns = QtWidgets.QAction(MainWindow)
        self.action2_5_pawns.setCheckable(True)
        self.action2_5_pawns.setObjectName("action2_5_pawns")
        self.action3_0_pawns = QtWidgets.QAction(MainWindow)
        self.action3_0_pawns.setCheckable(True)
        self.action3_0_pawns.setObjectName("action3_0_pawns")
        self.actionAvNone = QtWidgets.QAction(MainWindow)
        self.actionAvNone.setCheckable(True)
        self.actionAvNone.setChecked(False)
        self.actionAvNone.setObjectName("actionAvNone")
        self.action2_Halfmoves = QtWidgets.QAction(MainWindow)
        self.action2_Halfmoves.setCheckable(True)
        self.action2_Halfmoves.setObjectName("action2_Halfmoves")
        self.action4_Halfmoves = QtWidgets.QAction(MainWindow)
        self.action4_Halfmoves.setCheckable(True)
        self.action4_Halfmoves.setObjectName("action4_Halfmoves")
        self.action6_Halfmoves = QtWidgets.QAction(MainWindow)
        self.action6_Halfmoves.setCheckable(True)
        self.action6_Halfmoves.setObjectName("action6_Halfmoves")
        self.actionAll = QtWidgets.QAction(MainWindow)
        self.actionAll.setCheckable(True)
        self.actionAll.setObjectName("actionAll")
        self.actionConfigureEngine = QtWidgets.QAction(MainWindow)
        self.actionConfigureEngine.setObjectName("actionConfigureEngine")
        self.actionDebugEngine = QtWidgets.QAction(MainWindow)
        self.actionDebugEngine.setCheckable(True)
        self.actionDebugEngine.setObjectName("actionDebugEngine")
        self.actionAbout = QtWidgets.QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionPromote_Variant = QtWidgets.QAction(MainWindow)
        self.actionPromote_Variant.setObjectName("actionPromote_Variant")
        self.actionDelete_Variant = QtWidgets.QAction(MainWindow)
        self.actionDelete_Variant.setObjectName("actionDelete_Variant")
        self.actionAnnotate_All = QtWidgets.QAction(MainWindow)
        self.actionAnnotate_All.setObjectName("actionAnnotate_All")
        self.action1 = QtWidgets.QAction(MainWindow)
        self.action1.setCheckable(True)
        self.action1.setObjectName("action1")
        self.action2 = QtWidgets.QAction(MainWindow)
        self.action2.setCheckable(True)
        self.action2.setObjectName("action2")
        self.action3 = QtWidgets.QAction(MainWindow)
        self.action3.setCheckable(True)
        self.action3.setObjectName("action3")
        self.actionAll_2 = QtWidgets.QAction(MainWindow)
        self.actionAll_2.setObjectName("actionAll_2")
        self.actionUTF_8 = QtWidgets.QAction(MainWindow)
        self.actionUTF_8.setCheckable(True)
        self.actionUTF_8.setChecked(True)
        self.actionUTF_8.setObjectName("actionUTF_8")
        self.actionISO_8859_1 = QtWidgets.QAction(MainWindow)
        self.actionISO_8859_1.setCheckable(True)
        self.actionISO_8859_1.setObjectName("actionISO_8859_1")
        self.actionASCII = QtWidgets.QAction(MainWindow)
        self.actionASCII.setCheckable(True)
        self.actionASCII.setObjectName("actionASCII")
        self.actionHelp = QtWidgets.QAction(MainWindow)
        self.actionHelp.setObjectName("actionHelp")
        self.actionDemote_Variant = QtWidgets.QAction(MainWindow)
        self.actionDemote_Variant.setObjectName("actionDemote_Variant")
        self.actionPromote_Variant_to_Main = QtWidgets.QAction(MainWindow)
        self.actionPromote_Variant_to_Main.setObjectName("actionPromote_Variant_to_Main")
        self.actionSelect_Header_Elements = QtWidgets.QAction(MainWindow)
        self.actionSelect_Header_Elements.setObjectName("actionSelect_Header_Elements")
        self.actionRunning_Game = QtWidgets.QAction(MainWindow)
        self.actionRunning_Game.setObjectName("actionRunning_Game")
        self.menuRecent_PGN.addAction(self.actionbla)
        self.menuEncoding.addAction(self.actionUTF_8)
        self.menuEncoding.addAction(self.actionISO_8859_1)
        self.menuEncoding.addAction(self.actionASCII)
        self.menuFile.addAction(self.menuEncoding.menuAction())
        self.menuFile.addAction(self.actionOpen_PGN)
        self.menuFile.addAction(self.menuRecent_PGN.menuAction())
        self.menuFile.addAction(self.actionAppend_to_DB)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave_PGN)
        self.menuFile.addAction(self.actionSave_Game)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionExit)
        self.menuEdit.addAction(self.actionCopy_PGN)
        self.menuEdit.addAction(self.actionPaste_PGN)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionCopy_FEN)
        self.menuEdit.addAction(self.actionPaste_FEN)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionNext_Move)
        self.menuEdit.addAction(self.actionPrevious_Move)
        self.menuEdit.addAction(self.actionUndo_Last_Move)
        self.menuEdit.addSeparator()
        self.menuEdit.addAction(self.actionNext_Variant)
        self.menuEdit.addAction(self.actionPrevious_Variant)
        self.menuEdit.addAction(self.actionPromote_Variant)
        self.menuEdit.addAction(self.actionPromote_Variant_to_Main)
        self.menuEdit.addAction(self.actionDemote_Variant)
        self.menuEdit.addAction(self.actionDelete_Variant)
        self.menuEnd_Game.addAction(self.actionRunning_Game)
        self.menuEnd_Game.addAction(self.actionWhite_wins)
        self.menuEnd_Game.addAction(self.action0_1)
        self.menuEnd_Game.addAction(self.action1_2_1_2)
        self.menuGame.addAction(self.actionNew_Game)
        self.menuGame.addAction(self.actionSelect_Header_Elements)
        self.menuGame.addAction(self.menuEnd_Game.menuAction())
        self.menuGame.addSeparator()
        self.menuGame.addAction(self.actionFlip_Board)
        self.menuGame.addSeparator()
        self.menuGame.addAction(self.actionShow_Options)
        self.menuGame.addAction(self.actionWarn_of_Danger)
        self.menuSelect_Engine.addSeparator()
        self.menuSearch_Depth.addAction(self.action5_Moves)
        self.menuSearch_Depth.addAction(self.action10_Moves)
        self.menuSearch_Depth.addAction(self.action15_Moves)
        self.menuSearch_Depth.addAction(self.action20_Moves)
        self.menuSearch_Depth.addAction(self.action25_Moves)
        self.menuBlunder_Limit.addAction(self.actionBlNone)
        self.menuBlunder_Limit.addSeparator()
        self.menuBlunder_Limit.addAction(self.action0_5_pawns)
        self.menuBlunder_Limit.addAction(self.action1_pawn)
        self.menuBlunder_Limit.addAction(self.action1_5_pawns)
        self.menuBlunder_Limit.addAction(self.action2_0_pawns)
        self.menuBlunder_Limit.addAction(self.action2_5_pawns)
        self.menuBlunder_Limit.addAction(self.action3_0_pawns)
        self.menuAnnotate_Variants.addAction(self.actionAvNone)
        self.menuAnnotate_Variants.addAction(self.actionAll)
        self.menuAnnotate_Variants.addSeparator()
        self.menuAnnotate_Variants.addAction(self.action2_Halfmoves)
        self.menuAnnotate_Variants.addAction(self.action4_Halfmoves)
        self.menuAnnotate_Variants.addAction(self.action6_Halfmoves)
        self.menuNumber_of_Annotations.addAction(self.action1)
        self.menuNumber_of_Annotations.addAction(self.action2)
        self.menuNumber_of_Annotations.addAction(self.action3)
        self.menuEngine.addAction(self.menuSelect_Engine.menuAction())
        self.menuEngine.addAction(self.menuSearch_Depth.menuAction())
        self.menuEngine.addSeparator()
        self.menuEngine.addAction(self.actionAnnotate_Last_Move)
        self.menuEngine.addAction(self.actionAnnotate_All)
        self.menuEngine.addAction(self.menuNumber_of_Annotations.menuAction())
        self.menuEngine.addAction(self.menuBlunder_Limit.menuAction())
        self.menuEngine.addAction(self.menuAnnotate_Variants.menuAction())
        self.menuEngine.addSeparator()
        self.menuEngine.addAction(self.actionShow_Scores)
        self.menuEngine.addAction(self.actionShow_Hints)
        self.menuEngine.addSeparator()
        self.menuEngine.addAction(self.actionConfigureEngine)
        self.menuEngine.addAction(self.actionDebugEngine)
        self.menuDatabase.addAction(self.actionNew_PGN)
        self.menuDatabase.addAction(self.actionClose_PGN)
        self.menuDatabase.addSeparator()
        self.menuDatabase.addAction(self.actionAdd_Game)
        self.menuDatabase.addAction(self.actionRemove_Games)
        self.menuHelp.addAction(self.actionHelp)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionAbout)
        self.menuBar.addAction(self.menuFile.menuAction())
        self.menuBar.addAction(self.menuEdit.menuAction())
        self.menuBar.addAction(self.menuDatabase.menuAction())
        self.menuBar.addAction(self.menuGame.menuAction())
        self.menuBar.addAction(self.menuEngine.menuAction())
        self.menuBar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(2)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MzChess"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabGame), _translate("MainWindow", "Game"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabHeader), _translate("MainWindow", "Header"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabDatabase), _translate("MainWindow", "Database"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabScorePlot), _translate("MainWindow", "Score Plot"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabLog), _translate("MainWindow", "Log"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuRecent_PGN.setTitle(_translate("MainWindow", "Recent"))
        self.menuEncoding.setTitle(_translate("MainWindow", "Encoding"))
        self.menuEdit.setTitle(_translate("MainWindow", "Edit"))
        self.menuGame.setTitle(_translate("MainWindow", "Game"))
        self.menuEnd_Game.setTitle(_translate("MainWindow", "End Game"))
        self.menuEngine.setTitle(_translate("MainWindow", "Engine"))
        self.menuSelect_Engine.setTitle(_translate("MainWindow", "Select Engine"))
        self.menuSearch_Depth.setTitle(_translate("MainWindow", "Search Depth"))
        self.menuBlunder_Limit.setTitle(_translate("MainWindow", "Blunder Limit"))
        self.menuAnnotate_Variants.setTitle(_translate("MainWindow", "Annotate Variants"))
        self.menuNumber_of_Annotations.setTitle(_translate("MainWindow", "# Annotations"))
        self.menuDatabase.setTitle(_translate("MainWindow", "Database"))
        self.menuHelp.setTitle(_translate("MainWindow", "Help"))
        self.actionCopy_FEN.setText(_translate("MainWindow", "Copy FEN"))
        self.actionCopy_FEN.setShortcut(_translate("MainWindow", "Ctrl+Shift+C"))
        self.actionPaste_FEN.setText(_translate("MainWindow", "Paste FEN"))
        self.actionPaste_FEN.setShortcut(_translate("MainWindow", "Ctrl+Shift+V"))
        self.actionCopy_PGN.setText(_translate("MainWindow", "Copy PGN"))
        self.actionCopy_PGN.setShortcut(_translate("MainWindow", "Ctrl+C"))
        self.actionPaste_PGN.setText(_translate("MainWindow", "Paste PGN"))
        self.actionPaste_PGN.setShortcut(_translate("MainWindow", "Ctrl+V"))
        self.actionNew_Game.setText(_translate("MainWindow", "New"))
        self.actionWhite_wins.setText(_translate("MainWindow", "1-0"))
        self.action0_1.setText(_translate("MainWindow", "0-1"))
        self.action1_2_1_2.setText(_translate("MainWindow", "1/2-1/2"))
        self.actionShow_Options.setText(_translate("MainWindow", "Show Move Options"))
        self.action5_Moves.setText(_translate("MainWindow", "5 Moves"))
        self.action10_Moves.setText(_translate("MainWindow", "10 Moves"))
        self.action15_Moves.setText(_translate("MainWindow", "15 Moves"))
        self.action20_Moves.setText(_translate("MainWindow", "20 Moves"))
        self.action25_Moves.setText(_translate("MainWindow", "25 Moves"))
        self.actionShow_Hints.setText(_translate("MainWindow", "Show Hints"))
        self.actionNext_Move.setText(_translate("MainWindow", "Next Move"))
        self.actionNext_Move.setShortcut(_translate("MainWindow", "Down"))
        self.actionPrevious_Move.setText(_translate("MainWindow", "Previous Move"))
        self.actionPrevious_Move.setShortcut(_translate("MainWindow", "Up"))
        self.actionOpen_PGN.setText(_translate("MainWindow", "Open DB ..."))
        self.actionSave_PGN.setText(_translate("MainWindow", "Save DB ..."))
        self.actionExit.setText(_translate("MainWindow", "Exit"))
        self.actionbla.setText(_translate("MainWindow", "bla"))
        self.actionAdd_To_DB.setText(_translate("MainWindow", "Add to DB"))
        self.actionAppend_to_DB.setText(_translate("MainWindow", "Append to DB"))
        self.actionSave_Game.setText(_translate("MainWindow", "Save Game ..."))
        self.actionClose_PGN.setText(_translate("MainWindow", "Close"))
        self.actionAdd_Game.setText(_translate("MainWindow", "Add Game"))
        self.actionRemove_Games.setText(_translate("MainWindow", "Remove Games"))
        self.actionNew_PGN.setText(_translate("MainWindow", "New"))
        self.actionFlip_Board.setText(_translate("MainWindow", "Flip Board"))
        self.actionFlip_Board.setShortcut(_translate("MainWindow", "Ctrl+F"))
        self.actionNext_Variant.setText(_translate("MainWindow", "Next Variant"))
        self.actionNext_Variant.setShortcut(_translate("MainWindow", "Right"))
        self.actionPrevious_Variant.setText(_translate("MainWindow", "Previous Variant"))
        self.actionPrevious_Variant.setShortcut(_translate("MainWindow", "Left"))
        self.actionShow_Scores.setText(_translate("MainWindow", "Show Scores"))
        self.actionWarn_of_Danger.setText(_translate("MainWindow", "Warn of Danger"))
        self.actionWarn_of_Danger.setShortcut(_translate("MainWindow", "Ctrl+W"))
        self.actionUndo_Last_Move.setText(_translate("MainWindow", "Undo Last Move"))
        self.actionUndo_Last_Move.setShortcut(_translate("MainWindow", "Ctrl+Z"))
        self.actionAnnotate_Last_Move.setText(_translate("MainWindow", "Annotate Last Move"))
        self.actionBlNone.setText(_translate("MainWindow", "None"))
        self.action0_5_pawns.setText(_translate("MainWindow", "0.5 pawns"))
        self.action1_pawn.setText(_translate("MainWindow", "1 pawn"))
        self.action1_5_pawns.setText(_translate("MainWindow", "1.5 pawns"))
        self.action2_0_pawns.setText(_translate("MainWindow", "2 pawns"))
        self.action2_5_pawns.setText(_translate("MainWindow", "2.5 pawns"))
        self.action3_0_pawns.setText(_translate("MainWindow", "3 pawns"))
        self.actionAvNone.setText(_translate("MainWindow", "None"))
        self.action2_Halfmoves.setText(_translate("MainWindow", "2 Halfmoves"))
        self.action4_Halfmoves.setText(_translate("MainWindow", "4 Halfmoves"))
        self.action6_Halfmoves.setText(_translate("MainWindow", "6 Halfmoves"))
        self.actionAll.setText(_translate("MainWindow", "All"))
        self.actionConfigureEngine.setText(_translate("MainWindow", "Configure ..."))
        self.actionDebugEngine.setText(_translate("MainWindow", "Debug"))
        self.actionAbout.setText(_translate("MainWindow", "About ..."))
        self.actionPromote_Variant.setText(_translate("MainWindow", "Promote Variant"))
        self.actionPromote_Variant.setShortcut(_translate("MainWindow", "Ctrl+P"))
        self.actionDelete_Variant.setText(_translate("MainWindow", "Delete Variant"))
        self.actionDelete_Variant.setShortcut(_translate("MainWindow", "Ctrl+R"))
        self.actionAnnotate_All.setText(_translate("MainWindow", "Annotate All"))
        self.action1.setText(_translate("MainWindow", "1 variant"))
        self.action2.setText(_translate("MainWindow", "2 variants"))
        self.action3.setText(_translate("MainWindow", "3 variants"))
        self.actionAll_2.setText(_translate("MainWindow", "All"))
        self.actionUTF_8.setText(_translate("MainWindow", "UTF-8"))
        self.actionISO_8859_1.setText(_translate("MainWindow", "ISO 8859/1"))
        self.actionASCII.setText(_translate("MainWindow", "ASCII"))
        self.actionHelp.setText(_translate("MainWindow", "Help ..."))
        self.actionDemote_Variant.setText(_translate("MainWindow", "Demote Variant"))
        self.actionDemote_Variant.setShortcut(_translate("MainWindow", "Ctrl+D"))
        self.actionPromote_Variant_to_Main.setText(_translate("MainWindow", "Promote Variant to Main"))
        self.actionPromote_Variant_to_Main.setShortcut(_translate("MainWindow", "Ctrl+M"))
        self.actionSelect_Header_Elements.setText(_translate("MainWindow", "Select Header Elements ..."))
        self.actionRunning_Game.setText(_translate("MainWindow", "*"))
from gameheaderview import GameHeaderView
from gamelisttableview import GameListTableView
from gametreeview import GameTreeView
from qboardviewclass import QBoardViewClass
from scoreplotgraphicsview import ScorePlot
from uciedit import QUCIEdit


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Users\Reinh\OneDrive\Dokumente\python\MzChess\textEdit.ui'
#
# Created by: PyQt5 UI code generator 5.15.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(400, 300)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.plainTextEdit = QtWidgets.QPlainTextEdit(Dialog)
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.verticalLayout.addWidget(self.plainTextEdit)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.UnitedStates))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



import os.path
import PyQt5.QtCore,  PyQt5.QtGui, PyQt5.QtWidgets 

class HelpBrowser(PyQt5.QtWidgets.QWidget):
 indexFile = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'build', 'html', 'index.html')

 def __init__(self):
   super().__init__()
   
   self.tb = PyQt5.QtWidgets.QTextBrowser()
   self.tb.setAcceptRichText(True)
   self.tb.setOpenExternalLinks(True)
   self.tb.setOpenLinks(True)
   with open(self.indexFile, 'r',  encoding = 'utf-8-sig') as f:
    indexText = f.read()
   self.tb.setHtml(indexText)
   # self.tb.setSource(PyQt5.QtCore.QUrl(self.indexFile))

   vbox = PyQt5.QtWidgets.QVBoxLayout()
   vbox.addWidget(self.tb, 1)

   self.setLayout(vbox)

   self.setWindowTitle('QTextBrowser')
   self.setGeometry(100, 100, 800, 800)
   self.show()


if __name__ == '__main__':
 import sys

 app = PyQt5.QtWidgets.QApplication(sys.argv)
 # ex = HelpBrowser()
 PyQt5.QtGui.QDesktopServices.openUrl(PyQt5.QtCore.QUrl.fromLocalFile(HelpBrowser.indexFile))

 sys.exit(app.exec_())



'''
Configuration of Engines
================================

This dialog allows to add/remove/configure engines

|ConfigureEngine| 

Configuration of the Engine's Options
===========================================

By pressing the *Details* button the *Configure Engine Options* dialog is started.
Please consult the engine's documentation (e.g. `stockfish`_ or `here`_) before changing any options.

|ConfigureEngineOptions| 

.. |ConfigureEngine| image:: configureEngine.png
  :width: 400
  :alt: Configure Engine
.. |ConfigureEngineOptions| image:: configureEngineOptions.png
  :width: 400
  :alt: Configure Engine Options
.. _stockfish: https://github.com/official-stockfish/Stockfish
.. _here: https://www.chessprogramming.org/Engines
'''
from typing import List, Dict, Callable, Union, Optional, Any
import os, os.path
import platform
import copy

import configparser
import re

import PyQt5.QtCore
import PyQt5.QtGui
import PyQt5.QtWidgets

from chessengine import ChessEngine
from configureEngineOptions import ConfigureEngineOptions 

import Ui_configureEngine

intRe = re.compile(r"^[+\-]?[0-9]+$")
boolRe = re.compile(r"^(True|False)$")

def loadEngineSettings(settings : configparser.ConfigParser) -> Dict[str, List[Any]]:
 engineDict = dict()
 for section in settings.sections():
  partList = section.split('/')
  isEngineSection = (len(partList) == 2 and partList[0] == 'Engine')
  if  isEngineSection:
   key = partList[1]
   optionsDict = dict()
   executable = None
   for opt in settings[section]:
    value = settings[section][opt]
    if opt == 'executable':
     executable = value
    elif intRe.match(value) is not None:
     optionsDict[opt] = int(value)
    elif boolRe.match(value) is not None:
     optionsDict[opt] = bool(value)
    else:
     optionsDict[opt] = value
   if executable is not None:
    engineDict[key] = [executable, optionsDict]
 return engineDict

def saveEngineSettings(settings : configparser.ConfigParser, engineDict : Dict[str, List[Any]]) -> None:
 for section in settings.sections():
  partList = section.split('/')
  if len(partList) == 2 and partList[0] == 'Engine':
   settings.remove_section(section)
 for opt in engineDict:
  executable, optionsDict = engineDict[opt]
  key = 'Engine/{}'.format(opt)
  settings[key] = optionsDict
  settings[key]['executable'] = executable

class ConfigureEngine(PyQt5.QtWidgets.QDialog, Ui_configureEngine.Ui_Dialog):
 fileDialogOptions = PyQt5.QtWidgets.QFileDialog.Options() \
                        | PyQt5.QtWidgets.QFileDialog.DontUseNativeDialog
 fileDialogFilters = PyQt5.QtCore.QDir.AllDirs \
                      | PyQt5.QtCore.QDir.Files \
                      | PyQt5.QtCore.QDir.NoDotAndDotDot \
                      | PyQt5.QtCore.QDir.Hidden

 def __init__(self, parent : Optional[PyQt5.QtCore.QObject] = None):
  super(ConfigureEngine, self).__init__(parent)
  self.setupUi(self)
  self.engineDict = dict()
  self.log = False
  self.directories = list()
  self.engineTableWidget.horizontalHeader().setStretchLastSection(True)
  self.engineTableWidget.setSelectionBehavior(PyQt5.QtWidgets.QAbstractItemView.SelectRows)
  self.browsePushButton.setIcon(self.style().standardIcon(PyQt5.QtWidgets.QStyle.SP_DialogSaveButton))
  self.engineDetails = dict()

 @PyQt5.QtCore.pyqtSlot(str)
 def notifyError(self, msg):
  msgBox = PyQt5.QtWidgets.QMessageBox()
  msgBox.setIcon(PyQt5.QtWidgets.QMessageBox.Critical)
  msgBox.setText(msg)
  msgBox.setWindowTitle("Error")
  msgBox.exec()

 def _fillForm(self):
  self.engineTableWidget.setRowCount(len(self.engineDict))
  for row, n2ec in enumerate(self.engineDict.items()):
   name, ec = n2ec
   executable = ec[0]
   if not (os.path.isfile(executable) and os.access(executable, os.X_OK )):
    executable = 'not found'
   else:
    self.directories.append(os.path.dirname(executable))
   self.engineTableWidget.setCellWidget(row, 0, PyQt5.QtWidgets.QLabel(name))
   self.engineTableWidget.setCellWidget(row, 1, PyQt5.QtWidgets.QLabel(executable))
  self.executableLineEdit.setText('')
  self.nameLineEdit.setText('')

 def run(self, engineDict : Dict[str, List[Any]] = dict(),  log : Union[Callable, bool] = False) -> Dict[str, List[Any]]:
  self.engineDict = copy.deepcopy(engineDict)
  self.log = log
  self._fillForm()
  self.isChanged = False
  self.selectedRow = None
  if self.exec():
   return self.engineDict
  else:
   return engineDict
  
 def ignoreIsChanged(self) -> None:
  if self.isChanged:
   msgBox = PyQt5.QtWidgets.QMessageBox()
   msgBox.setIcon(PyQt5.QtWidgets.QMessageBox.Critical)
   msgBox.addButton(PyQt5.QtWidgets.QMessageBox.No)
   msgBox.addButton(PyQt5.QtWidgets.QMessageBox.Yes)
   msgBox.setText('Ignore ?')
   msgBox.setWindowTitle("Unsaved content detected")
   rc = msgBox.exec()
   return rc
  else:
   return True

 def checkEngine(self, executable : str) -> Optional[ChessEngine]:
  try:
   engine = ChessEngine([executable, dict()], log = self.log)
   isOK = engine.isReady()
  except:
   isOK = False
  if not isOK:
   msgBox = PyQt5.QtWidgets.QMessageBox()
   msgBox.setIcon(PyQt5.QtWidgets.QMessageBox.Critical)
   msgBox.addButton(PyQt5.QtWidgets.QMessageBox.Ok)
   msgBox.setText('{} is not an UCI engine'.format(executable))
   msgBox.setWindowTitle("Error")
   msgBox.exec()
   return None
  return engine
 
 @PyQt5.QtCore.pyqtSlot()
 def on_detailsPushButton_clicked(self):
  executable = self.executableLineEdit.text()   
  if len(executable) == 0:
   self.notifyError('Executable undefined')
   return
  name = self.nameLineEdit.text()
  engine = self.checkEngine(executable)
  if engine is not None:
   dialog = ConfigureEngineOptions()
   newEngineDetails = dialog.run(engine.optionsDict, changedOptions = self.engineDetails, name = name)
   if newEngineDetails != self.engineDetails:
    self.engineDetails = newEngineDetails
    self.isChanged = True

 @PyQt5.QtCore.pyqtSlot()
 def on_browsePushButton_clicked(self):
  fDialog = PyQt5.QtWidgets.QFileDialog()
  fDialog.setFileMode(PyQt5.QtWidgets.QFileDialog.ExistingFile)
  fDialog.setOptions(self.fileDialogOptions)
  fDialog.setFilter(self.fileDialogFilters)
  fDialog.setWindowTitle("Load Engine Executable ...")
  if len(self.directories) > 0:
   fDialog.setHistory(self.directories)
  if len(self.executableLineEdit.text()) > 0:
   head, tail = os.path.split(self.executableLineEdit.text())
   fDialog.setDirectory(head)
   fDialog.selectFile(tail)
  if platform.system() == 'Windows':
   fDialog.setNameFilter("Executables (*.exe);;All Files (*)")
  else:
   fDialog.setNameFilter("All Files (*)")
  if fDialog.exec():
   executable = fDialog.selectedFiles()[0]
   self.engineDetails = dict()
   if self.checkEngine(executable) is not None:
    self.executableLineEdit.setText(executable)
    self.isChanged = True

 @PyQt5.QtCore.pyqtSlot(int, int)
 def on_engineTableWidget_cellClicked(self, row, column):
  if not self.ignoreIsChanged():
   return  
  name = self.engineTableWidget.cellWidget(row, 0).text()
  self.nameLineEdit.setText(name)
  self.executableLineEdit.setText(self.engineDict[name][0])
  self.engineDetails = self.engineDict[name][1]
  self.selectedRow = row

 @PyQt5.QtCore.pyqtSlot(str)
 def on_nameLineEdit_textEdited(self, newName):
  self.isChanged = True

 @PyQt5.QtCore.pyqtSlot(str)
 def on_executableLineEdit_textEdited(self, newExecutable):
  if self.checkEngine(newExecutable) is None:
   self.executableLineEdit.setText('')
   self.engineDetails = dict()
  self.isChanged = True

 @PyQt5.QtCore.pyqtSlot()
 def on_savePushButton_clicked(self):
  name = self.nameLineEdit.text()
  executable = self.executableLineEdit.text()
  if not self.isChanged:
   return
  if len(executable) == 0:
   self.notifyError('Executable undefined')
   return
  self.engineDict[name] = [executable, self.engineDetails]
  self._fillForm()
  self.isChanged = False
 
 @PyQt5.QtCore.pyqtSlot()
 def on_removePushButton_clicked(self):
  if not self.ignoreIsChanged():
   return
  key = self.engineTableWidget.cellWidget(self.selectedRow, 0).text()
  self.engineDict.pop(key, None)
  self.engineTableWidget.removeRow(self.selectedRow)
  self.selectedRow = None
  
 @PyQt5.QtCore.pyqtSlot()
 def on_cancelPushButton_clicked(self):
  self.done(False)
  
 @PyQt5.QtCore.pyqtSlot()
 def on_okPushButton_clicked(self):
  if self.ignoreIsChanged():
   self.done(True)

if __name__ == "__main__":
 import os,  os.path
 import sys
 import configparser

 qApp=PyQt5.QtWidgets.QApplication(sys.argv)   

 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 os.chdir(fileDirectory)
 
 settingsFile = os.path.join(fileDirectory, 'settings.ini')
 settings = configparser.ConfigParser(delimiters=['='], allow_no_value=True)
 settings.optionxform = str
 settings.read(settingsFile, encoding = 'utf-8')
 engineDict = loadEngineSettings(settings)

 configureEngine = ConfigureEngine()
 print(engineDict)
 for name in engineDict:
  engineDict[name] = [engineDict[name], dict()] 
 configureEngine.show()
 newEngineDict = configureEngine.run(engineDict = engineDict, log = True)
 print('------- new')
 print(newEngineDict)
 qApp.exec_()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Users\Reinh\OneDrive\Dokumente\python\MzChess\configureEngine.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(634, 301)
        Dialog.setSizeGripEnabled(True)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Dialog)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)
        self.cancelPushButton = QtWidgets.QPushButton(Dialog)
        self.cancelPushButton.setMinimumSize(QtCore.QSize(75, 0))
        self.cancelPushButton.setMaximumSize(QtCore.QSize(75, 16777215))
        self.cancelPushButton.setObjectName("cancelPushButton")
        self.gridLayout.addWidget(self.cancelPushButton, 4, 4, 1, 1)
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.removePushButton = QtWidgets.QPushButton(Dialog)
        self.removePushButton.setMinimumSize(QtCore.QSize(75, 0))
        self.removePushButton.setMaximumSize(QtCore.QSize(75, 16777215))
        self.removePushButton.setObjectName("removePushButton")
        self.gridLayout.addWidget(self.removePushButton, 4, 1, 1, 1)
        self.browsePushButton = QtWidgets.QPushButton(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.browsePushButton.sizePolicy().hasHeightForWidth())
        self.browsePushButton.setSizePolicy(sizePolicy)
        self.browsePushButton.setMinimumSize(QtCore.QSize(23, 0))
        self.browsePushButton.setMaximumSize(QtCore.QSize(23, 16777215))
        self.browsePushButton.setText("")
        self.browsePushButton.setObjectName("browsePushButton")
        self.gridLayout.addWidget(self.browsePushButton, 1, 6, 1, 1)
        self.nameLineEdit = QtWidgets.QLineEdit(Dialog)
        self.nameLineEdit.setMinimumSize(QtCore.QSize(0, 23))
        self.nameLineEdit.setObjectName("nameLineEdit")
        self.gridLayout.addWidget(self.nameLineEdit, 0, 1, 1, 5)
        self.engineTableWidget = QtWidgets.QTableWidget(Dialog)
        self.engineTableWidget.setEditTriggers(QtWidgets.QAbstractItemView.CurrentChanged|QtWidgets.QAbstractItemView.SelectedClicked)
        self.engineTableWidget.setWordWrap(False)
        self.engineTableWidget.setColumnCount(2)
        self.engineTableWidget.setObjectName("engineTableWidget")
        self.engineTableWidget.setRowCount(0)
        self.engineTableWidget.horizontalHeader().setVisible(False)
        self.gridLayout.addWidget(self.engineTableWidget, 3, 0, 1, 6)
        self.savePushButton = QtWidgets.QPushButton(Dialog)
        self.savePushButton.setMinimumSize(QtCore.QSize(75, 0))
        self.savePushButton.setMaximumSize(QtCore.QSize(75, 16777215))
        self.savePushButton.setObjectName("savePushButton")
        self.gridLayout.addWidget(self.savePushButton, 4, 0, 1, 1)
        self.detailsPushButton = QtWidgets.QPushButton(Dialog)
        self.detailsPushButton.setObjectName("detailsPushButton")
        self.gridLayout.addWidget(self.detailsPushButton, 1, 5, 1, 1)
        self.okPushButton = QtWidgets.QPushButton(Dialog)
        self.okPushButton.setMinimumSize(QtCore.QSize(75, 0))
        self.okPushButton.setMaximumSize(QtCore.QSize(75, 16777215))
        self.okPushButton.setObjectName("okPushButton")
        self.gridLayout.addWidget(self.okPushButton, 4, 5, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 4, 2, 1, 2)
        self.executableLineEdit = QtWidgets.QLineEdit(Dialog)
        self.executableLineEdit.setObjectName("executableLineEdit")
        self.gridLayout.addWidget(self.executableLineEdit, 1, 1, 1, 4)
        self.horizontalLayout.addLayout(self.gridLayout)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Configure Engine"))
        self.label_2.setText(_translate("Dialog", "Executable:"))
        self.cancelPushButton.setText(_translate("Dialog", "Cancel"))
        self.label.setText(_translate("Dialog", "Name:"))
        self.removePushButton.setText(_translate("Dialog", "Remove"))
        self.savePushButton.setText(_translate("Dialog", "Save"))
        self.detailsPushButton.setText(_translate("Dialog", "Details"))
        self.okPushButton.setText(_translate("Dialog", "OK"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



__description__ = \
"""
A PyQt5 based chess GUI

Versions:
 1.0.0      first version
 1.0.3      installation running
 1.0.4      bug fixes + end-of-game handling completed
"""
__author__ = "Reinhard Maerz"
__date__ = "2021-07-11"
__version__ = "1.0.5"

import os
import sys

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

__all__ = [
 'AboutDialog', 
 'AnnotateEngine', 'Annotator', 
 'ChessMainWindow', 
 'ChessEngine', 
 'ConfigureEngine', 'loadEngineSettings', 'saveEngineSettings', 
 'ConfigureEngineOptions',  
 'ECODatabase', 'TSVType', 
 'GameHeaderView', 'KeyType', 
 'GameListTableModel', 'GameListTableView', 
 'HelpBrowser', 
 'read_game', 'read_board', 'read_headers', 'skip_game', 'PGNLexer', 
 'ScorePlot', 
 'ButtonLine', 'ItemSelector', 'treeWidgetItemPos', 
 'QUCIEdit', 'UCIHighlighter', 
 'warnOfDanger'
]

from .AboutDialog import AboutDialog
from .annotateEngine import AnnotateEngine,  Annotator
from .chessMainWindow import ChessMainWindow
from .chessengine import ChessEngine
from .configureEngine import ConfigureEngine, loadEngineSettings, saveEngineSettings
from .configureEngineOptions import ConfigureEngineOptions
from .eco import ECODatabase, TSVType
from .gameheaderview import GameHeaderView, KeyType
from .gamelisttableview import GameListTableModel, GameListTableView
from .helpDialog import HelpBrowser
from .pgnParse import read_game, read_board, read_headers, skip_game, PGNLexer
from .qboardviewclass import QBoardViewClass, Piece, Game
from .scoreplotgraphicsview import ScorePlot
from .specialDialogs import ButtonLine, ItemSelector, treeWidgetItemPos
from .uciedit import QUCIEdit, UCIHighlighter
from .warnOfDanger import warnOfDanger



'''A Wrapper for Encyclopaedia of Chess Openings (ECO) (see `fiekas.eco`_)
It makes use of `polyglot`_ books to arrange the games.
 
.. _fiekas.eco: https://github.com/niklasf/eco
.. _polyglot: https://sourceforge.net/projects/codekiddy-chess/files/Books/Polyglot%20books/
'''

from typing import Optional, Union, Dict, List, Any
import os, os.path
import glob
import warnings
from enum import IntEnum, unique

import chess, chess.pgn, chess.polyglot

@unique
class TSVType(IntEnum):
 'Columns of TSV-files'
 ECO = 0
 OPENING = 1
 FEN = 2
 MOVES = 3

class ECODatabase(list):
 '''A Wrapper class for Encyclopaedia of Chess Openings (ECO)

:param ecoDirectory: directory of tsv - files
:param tsvPattern: a glob pattern describing the files to be loaded. ``None`` -> nothing is loaded
 '''
 keys =  ['eco', 'name', 'fen', 'moves']
 
 def __init__(self, 
  ecoDirectory : os.PathLike = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'eco'),
  tsvPattern : Optional[str] = '*.tsv') -> None:
  super(ECODatabase, self).__init__()
  self.ecoDirectory = os.path.abspath(ecoDirectory)
  if tsvPattern is None or len(tsvPattern) == 0:
   return
  for tsvFile in glob.glob(os.path.join(self.ecoDirectory, tsvPattern)):
   self.loadTSVFile(tsvFile)
   # print('{} lines read from file {}'.format(len(self), os.path.basename(tsvFile)))

 def loadTSVFile(self, tsvFile : os.PathLike) -> None:
  '''Adds a *tsvFile* to the database 

:param tsvFile: tsv - file to be added
  '''
  if not os.path.isabs(tsvFile):
   tsvFile = os.path.join(self.ecoDirectory, tsvFile)
  if not os.path.isfile(tsvFile):
   raise IOError('ECODatabase/loadTSVFile: ECO file {} not found'.format(tsvFile))
  with open(tsvFile, mode='r') as f:
   isFirst = True
   while True:
    row = f.readline().strip(' \n').split('\t')
    if isFirst:
     if row != self.keys:
      raise IOError('ECODatabase/loadTSVFile: {} is not a valid ECO file'.format(tsvFile))
     isFirst = False
     continue
    if len(row) != 4:
     f.close()
     return
    moveList = list()
    for uciCode in row[3].split(' '):
     moveList.append(chess.Move.from_uci(uciCode))
    self.append((row[0], row[1], row[2], moveList))
 
 def column2IdList(self, key : TSVType) -> Dict[str, List[int]]:
  resultDict = dict()
  for id, tpl in enumerate(self):
   _key = tpl[key]
   if _key not in resultDict:
    resultDict[_key] = list()
   resultDict[_key].append(id)
  return resultDict

 def column2Id(self, column : int) -> Dict[str, int]:
  # standard databases => only fen is unique 
  assert column >= 0 and column < 3
  resultDict = dict()
  for id, tpl in enumerate(self):
   key = tpl[column]
   if key in resultDict:
    warnings.warn('ECODatabase/column2Id: Duplicate key {}'.format(key), RuntimeWarning)
   resultDict[key] = id
  return resultDict
  
 def fen2Id(self) -> Dict[str, int]:
  return self.column2Id(2)

 def column2Count(self, column) -> Dict[str, int]:
  assert column >= 0 and column < 3
  resultDict = dict()
  for tpl in self:
   key = tpl[column]
   if key not in resultDict:
    resultDict[key] = 0
   resultDict[key] += 1
  return resultDict
  
 @staticmethod
 def commonLength(listOLists : List[List[Any]]) -> int:
  n = 0
  while True:
   itemSet = set()
   for valueList in listOLists:
    if len(valueList) < n + 1:
     return n
    itemSet.add(valueList[n])
   if len(itemSet) > 1:
    return n
   n += 1
  return 0
  
 def statistics(self, key : TSVType = TSVType.ECO) -> Dict[str, Dict[str, Union[int, List[int]]]]:
  '''Delivers statistical data. ``key`` one of the 4 columns of a TSV-file, i.e.
  
  * *ECO* : eco code
  * *OPENING* : opening name
  * *FEN* : Forsyth-Edwards-Notation of the starting position
  * *MOVES* : move list

 The values of the returned dictionary are
 
  * *idList* : list of entries in the TSV table
  * *nItems* : len(idlist)
  * *nCommon* : number of common moves
  
:param key: column of a TSV-file (see above)
:returns: return dictionary
  '''
  statisticsDict = dict()
  for columnValue, idList in self.column2IdList(key).items():
   newDict = dict()
   newDict['idList'] = idList
   newDict['nItems'] = len(idList)
   listOfMoveLists = list()
   for id in idList:
     listOfMoveLists.append(self[id][3])
   newDict['nCommon'] = self.commonLength(listOfMoveLists)
   statisticsDict[columnValue] = newDict
  return statisticsDict
 
 @staticmethod
 def sortGameNode(gameNode : chess.pgn.GameNode):
  if len(gameNode.variations) == 0:
   return 1
  if len(gameNode.variations) == 1:
   return ECODatabase.sortGameNode(gameNode.variations[0]) + 1
  nItems2idList = list()
  for n in range(len(gameNode.variations)):
   nItems2idList.append((ECODatabase.sortGameNode(gameNode.variations[n]), n))
  nItems2idList = sorted(nItems2idList, reverse = True, key = lambda el : el[0])
  newVariations = list()
  for nItems, id in nItems2idList:
   newVariations.append(gameNode.variations[id])
  gameNode.variations = newVariations
  return nItems2idList[0][0] + 1
  
 def _createVariation(self, gameNode : chess.pgn.GameNode, nPly : int, idList : List[int], polyglotBook : Optional[chess.polyglot.MemoryMappedReader] = None) -> None:
  move2idDict = dict()
  for id in idList:
   actMoveList = self[id][3]
   if len(actMoveList) == nPly:
    gameNode.comment = '{} : {}'.format(self[id][0], self[id][1])
   else:
    actMove = actMoveList[nPly]
    if actMove not in move2idDict:
     move2idDict[actMove] = list()
    move2idDict[actMove].append(id)
  if polyglotBook is not None:
   w2mDict = dict()
   for entry in polyglotBook.find_all(gameNode.board()):
    if entry.move in move2idDict.keys():
     w2mDict[entry.weight] = entry.move
   moveList = list()
   for weight in sorted(w2mDict.keys(),  reverse = True):
    moveList.append(w2mDict[weight])
   for move in move2idDict.keys():
    if move not in moveList:
     moveList.append(move)
  else:
   moveList = move2idDict.keys()
  for move in moveList:
   newIdList = move2idDict[move]
   newGameNode = gameNode.add_variation(move)
   self._createVariation(newGameNode, nPly+1, newIdList)
  
 def createGame(self, listOfFirstMoves : List[chess.Move], polyglotBook : Optional[chess.polyglot.MemoryMappedReader] = None) -> chess.pgn.Game:
  '''Creates a opening game starting with ``listOfFirstMoves``. The opening variants are implemented as variants
    
:param listOfFirstMoves: list of starting moves
:param polyglotBook: if present, the moves are sorted with respect to frequency of use
:returns: return dictionary
  '''
  title = 'Openings starting with '
  gameNode = chess.pgn.Game()
  for move in listOfFirstMoves:
   title += ' {}'.format(chess.square_name(move.to_square))
   gameNode = gameNode.add_variation(move)
  game = gameNode.game()
  game.comment = title
  game.headers['Site'] = 'ECO Tables'
  game.headers['White'] = 'Niklas Fiekas/White'
  game.headers['Black'] = 'Niklas Fiekas/Black'
  nFirst = len(listOfFirstMoves)
  idList = list()
  for id, tpl in enumerate(self):
   actMoves = tpl[3]
   if len(actMoves) >= nFirst and actMoves[:nFirst] == listOfFirstMoves:
    idList.append(id)
  self._createVariation(gameNode, nFirst, idList, polyglotBook)
  return game
    
if __name__ == "__main__":
 def createOpeningGame(opening, moveList):
  global fileDirectory, eco, polyglotBook
  game = eco.createGame(moveList, polyglotBook = polyglotBook)
  newPgnName = os.path.join(fileDirectory, 'training','openings', '{}.pgn'.format(opening))
  ECODatabase.sortGameNode(game)
  # print(game)
  exporter = chess.pgn.StringExporter(headers=True, variations=True, comments=True)
  f = open(newPgnName, "w", encoding="utf-8")
  pgnString = game.accept(exporter)
  f.write(pgnString)
  f.close()
  print('{} bytes written to {}'.format(len(pgnString), newPgnName))

 fileDirectory = os.path.dirname(os.path.abspath(__file__))
 bookFile = os.path.join(fileDirectory,'books','elo2400.bin')
 polyglotBook = chess.polyglot.open_reader(bookFile)
 eco = ECODatabase()
 fen2IdDict = eco.fen2Id()
 statistics = eco.statistics()
 for ecoCode in sorted(statistics):
  iDict = statistics[ecoCode] 
  print('{:3} : {:3}, {} items'.format(ecoCode, iDict['nCommon'], iDict['nItems'] ))

 createOpeningGame('kingsPawn', [chess.Move.from_uci('e2e4'), chess.Move.from_uci('e7e5')])
 createOpeningGame('sicilianDefense', [chess.Move.from_uci('e2e4'), chess.Move.from_uci('c7c5')])
 createOpeningGame('frenchDefense', [chess.Move.from_uci('e2e4'), chess.Move.from_uci('e7e6')])
 createOpeningGame('caroKannDefense', [chess.Move.from_uci('e2e4'), chess.Move.from_uci('c7c6')])
 createOpeningGame('pircDefense', [chess.Move.from_uci('e2e4'), chess.Move.from_uci('d7d6')])
 createOpeningGame('queensGambit', [chess.Move.from_uci('d2d4'), chess.Move.from_uci('d7d5')])
 createOpeningGame('indianDefense', [chess.Move.from_uci('d2d4'), chess.Move.from_uci('g8f6')])
 createOpeningGame('englishOpening', [chess.Move.from_uci('c2c4')])
 createOpeningGame('retiOpening', [chess.Move.from_uci('g7f6')])
